# This file is part of aiootp, an asynchronous one-time-pad based crypto
# and anonymity library.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2020 Gonzo Investigatory Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2020 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


__all__ = ["keygens", "AsyncKeys", "Keys"]


__doc__ = """
A collection of highlevel tools for creating & managing symmetric keys.
"""


from .asynchs  import *
from .commons import primes
from .commons import commons
from .randoms import csprng
from .randoms import acsprng
from .ciphers import keys
from .ciphers import akeys
from .ciphers import subkeys
from .ciphers import asubkeys
from .ciphers import OneTimePad
from .ciphers import keypair_ratchets
from .ciphers import akeypair_ratchets
from .generics import is_iterable
from .generics import comprehension
from .generics import sha_256_hmac
from .generics import asha_256_hmac
from .generics import MemberFromStaticMethod


@comprehension()
async def atable_key_gen(key=None, table=commons.ASCII_TABLE):
    """
    This table based key generator function converts any key string
    containing an arbitrary set of characters, into another key string
    containing the set of characters provided by the table argument.
    This generator produces the elements one at a time until the size
    limit has been reached, or produces an unending stream of elements
    if no size is specified.

    The ASCII_TABLE that's provided as a default, is a comprehensive set
    of ascii characters that are legible, unique, and have single octet
    byte representations. This ensures keys can be converted into a
    cryptographically secure byte stream of binary numbers from 32 to
    126.

    This generator function provides either deterministic keys from a
    user key, or generates a random sha_512 hash key and derives a
    random key with the desired table elements from this hash. The
    result is a random, normal distribution of characters from among the
    items within the table. The table argument can also contain an
    aribtrary set of objects. In this case, this generator function is
    essentially analogous to one that generates a series of random
    choices from the table, with the added functionality of allowing
    deterministic outputs that coincide with the key and size arguments.

    Usage Examples:

    # To produce a 60 byte key of characters from the default table
    key = "hotdiggitydog_thischowisyummy"
    new_key = ""
    async for char in atable_key_gen(key=key, size=60):
        new_key += char
    print(new_key)
    >>> keVDn~;NrEu/or9#KoXnncfNN5q;o"R!VeuQB=?fgd:0"+-C`~_LlpoOBzqR
    """
    if key == None:
        key = await acsprng(None)
    if not is_iterable(table):
        raise TypeError("table is not iterable")
    elif isinstance(table, dict):
        table = list(table.keys())
    prime = primes[32][0]
    table_size = len(table)
    async for index in akeys(key, key).aresize(16).aint(16):
        yield table[index % prime % table_size]


@comprehension()
def table_key_gen(key=None, table=commons.ASCII_TABLE):
    """
    This table based key generator function converts any key string
    containing an arbitrary set of characters, into another key string
    containing the set of characters provided by the table argument.
    This generator produces the characters one at a time until the size
    limit has been reached.

    The ASCII_TABLE that's provided as a default, is a comprehensive set
    of ascii characters that are legible, unique, and have single octet
    byte representations. This ensures keys can be converted into a
    cryptographically secure byte stream of binary numbers from 32 to
    126.

    This generator function provides either deterministic keys from a
    user key, or generates a random sha_512 hash key and derives a
    random key with the desired table elements from this hash. The
    result is a random, normal distribution of characters from among the
    items within the table. The table argument can also contain an
    aribtrary set of objects. In this case, this generator function is
    essentially analogous to one that generates a series of random
    choices from the table, with the added functionality of allowing
    deterministic outputs that coincide with the key and size arguments.

    The size parameter determines the number of bytes/elements the
    output will contain.

    Usage Example:
    key = "hotdiggitydog_thischowisyummy"
    new_key = ""
    for char in table_key_gen(key=key, size=60):
        new_key += char
    print(new_key)
    >>> keVDn~;NrEu/or9#KoXnncfNN5q;o"R!VeuQB=?fgd:0"+-C`~_LlpoOBzqR
    """
    if key == None:
        key = csprng(None)
    if not is_iterable(table):
        raise TypeError("table is not iterable")
    elif isinstance(table, dict):
        table = list(table.keys())
    prime = primes[32][0]
    table_size = len(table)
    for index in keys(key, key).resize(16).int(16):
        yield table[index % prime % table_size]


async def atable_key(key=None, table=commons.ASCII_TABLE, size=64):
    """
    This table based key function converts any key string containing
    any arbitrary set of characters, into another key string containing
    the set of items provided by the table argument.

    The ASCII_TABLE that's provided as a default, is a comprehensive set
    of ascii characters that are all legible, with unique, single octet
    byte representations. This ensures keys can be converted into byte
    streams of binary numbers from 32 to 126, with no duplicate values.

    This function provides either deterministic keys from a user key,
    or generates a random sha_512 hash key and derives a random key with
    the desired table elements from this hash. The result is a random,
    normal distribution of characters from among the items within
    the table. The table argument can also contain an aribtrary set of
    objects. In this case, this function is essentially analogous to
    one that generates list of random choices from the table, with the
    added functionality of allowing deterministic outputs that coincide
    with the key and size arguments.

    Usage Examples:

    key = "smellycaaaat, smelly caaaaat!"
    new_key = table_key(key=key, table="0123456789abcdef")
    print(new_key)
    >>> 4f271c61b0e615a7d3e9ac0161497034d047d4ecddc650ae054f829b3416818c

    new_key = table_key(key=key, size=len(key))
    print(new_key)
    >>> #mE)bOQD@lY%]Qwpb9Zi^32]jteVg
    """
    async with atable_key_gen(key=key, table=table)[:size] as generator:
        try:
            return await generator.ajoin()
        except:
            return await generator.alist()


def table_key(key=None, table=commons.ASCII_TABLE, size=64):
    """
    This table based key function converts any key string containing
    any arbitrary set of characters, into another key string containing
    the set of items provided by the table argument.

    The ASCII_TABLE that's provided as a default, is a comprehensive set
    of ascii characters that are all legible, with unique, single octet
    byte representations. This ensures keys can be converted into byte
    streams of binary numbers from 32 to 126, with no duplicate values.

    This function provides either deterministic keys from a user key,
    or generates a random sha_512 hash key and derives a random key with
    the desired table elements from this hash. The result is a random,
    normal distribution of characters from among the items within
    the table. The table argument can also contain an aribtrary set of
    objects. In this case, this function is essentially analogous to
    one that generates list of random choices from the table, with the
    added functionality of allowing deterministic outputs that coincide
    with the key and size arguments.

    The size parameter determines the number of bytes/elements the
    output will contain.


    Usage Examples:
    key = "smellycaaaat, smelly caaaaat!"
    new_key = table_key(key=key, table="0123456789abcdef")
    print(new_key)
    >>> 4f271c61b0e615a7d3e9ac0161497034d047d4ecddc650ae054f829b3416818c

    new_key = table_key(key=key, size=len(key))
    print(new_key)
    >>> #mE)bOQD@lY%]Qwpb9Zi^32]jteVg
    """
    with table_key_gen(key=key, table=table)[:size] as generator:
        try:
            return generator.join()
        except:
            return generator.list()


async def akeypair(entropy=csprng()):
    """
    Returns a pair of symmetric 512-bit hexidecimal keys from our fast
    cryptographically secure pseudo-random number generator.
    """
    return await acsprng(entropy), await acsprng(entropy)


def keypair(entropy=csprng()):
    """
    Returns a pair of symmetric 512-bit hexidecimal keys from our fast
    cryptographically secure pseudo-random number generator.
    """
    return csprng(entropy), csprng(entropy)


class Keys:
    """
    This simple class coordinates and manages a pair of symmetric keys
    for establishing an arbitrary number of secure, deterministic
    streams of key material through instances' ``__getitem__``, ``keys``
    & ``subkeys`` methods. The class also contains static method key
    generators which function independantly from instance states.
    """
    seed = csprng
    keys = staticmethod(keys)
    subkeys = staticmethod(subkeys)
    keypair = staticmethod(keypair)
    table_key = staticmethod(table_key)
    table_key_gen = staticmethod(table_key_gen)
    keypair_ratchets = staticmethod(keypair_ratchets)

    def __init__(self, key=None):
        """
        Creates a symmetric key pair used to create deterministic streams
        of key material.
        """
        self.key = key
        self.reset(both=False if key else True)

    def __getitem__(self, salt=""):
        """
        Provides a simple interface for users to create deterministic
        & externally uncorrelatable key material from a name or ``salt``.
        """
        return self.keys(key=self.key, salt=self.salt, pid=salt)

    def hmac(self, data=None, key=None, *, hasher=sha_256_hmac):
        """
        Creates an HMAC code of ``data`` using ``key`` & the hashing
        function ``hasher``.
        """
        return hasher(data, key=key if key else self.key)

    def test_hmac(
        self, data=None, key=None, hmac=None, *, hasher=sha_256_hmac
    ):
        """
        Tests the ``hmac`` code against the derived HMAC of ``data``
        using ``key`` & the hashing function ``hasher``.
        """
        if hmac == self.hmac(
            data=data, key=key if key else self.key, hasher=hasher
        ):
            return True
        else:
            raise ValueError("HMAC of ``data`` isn't valid.")

    def state(self):
        """
        Returns both the main ``self.key`` & the ephemeral ``self.salt``.
        """
        return self.key, self.salt

    def reset(self, both=True):
        """
        Creates a new pair of symmetric keys if ``both`` is truthy, else
        just updates the ephemeral ``self.salt`` key.
        """
        if both:
            self.key, self.salt = keypair()
        else:
            self.salt = csprng(self.seed())
        method_tools = MemberFromStaticMethod()
        for method in [keys, subkeys]:
            method_tools.convert(
                self, method.__name__, method, key=self.key, salt=self.salt
            )


class AsyncKeys:
    """
    This simple class coordinates and manages a pair of symmetric keys
    for establishing an arbitrary number of secure, deterministic
    streams of key material through its ``__getitem__``, ``akeys``,
    & ``asubkeys`` methods. The class also contains static method key
    generators which function independantly from instance states.
    """
    aseed = acsprng
    akeys = staticmethod(akeys)
    asubkeys = staticmethod(asubkeys)
    akeypair = staticmethod(akeypair)
    atable_key = staticmethod(atable_key)
    atable_key_gen = staticmethod(atable_key_gen)
    akeypair_ratchets = staticmethod(akeypair_ratchets)

    def __init__(self, key=None):
        """
        Creates a symmetric key pair used to create deterministic streams
        of key material.
        """
        self.key, self.salt = key, csprng() if key else keypair()
        method_tools = MemberFromStaticMethod()
        for method in [keys, subkeys]:
            method_tools.convert(
                self, method.__name__, method, key=self.key, salt=self.salt
            )

    def __getitem__(self, salt=""):
        """
        Provides a simple interface for users to create deterministic
        & externally uncorrelatable key material from a name or ``salt``.
        """
        return self.akeys(key=self.key, salt=self.salt, pid=salt)

    async def ahmac(
        self, data=None, key=None, *, hasher=asha_256_hmac
    ):
        """
        Creates an HMAC code of ``data`` using ``key`` & the hashing
        function ``hasher``.
        """
        return await hasher(data, key=key if key else self.key)

    async def atest_hmac(
        self, data=None, key=None, hmac=None, *, hasher=sha_256_hmac
    ):
        """
        Tests the ``hmac`` code against the derived HMAC of ``data``
        using ``key`` & the async hashing function ``hasher``.
        """
        if hmac == await self.ahmac(
            data=data, key=key if key else self.key, hasher=hasher
        ):
            return True
        else:
            raise ValueError("HMAC of ``data`` isn't valid.")

    async def astate(self):
        """
        Returns both the main ``self.key`` & the ephemeral ``self.salt``.
        """
        return self.key, self.salt

    async def areset(self, both=True):
        """
        Creates a new pair of symmetric keys if ``both`` is truthy, else
        just updates the ephemeral ``self.salt`` key.
        """
        if both:
            self.key, self.salt = await akeypair()
        else:
            self.salt = await acsprng(await self.aseed())
        method_tools = MemberFromStaticMethod()
        for method in [keys, subkeys]:
            method_tools.convert(
                self, method.__name__, method, key=self.key, salt=self.salt
            )


async def ainsert_keyrings(self, key=None):
    """
    A generic __init__ function that can be copied into abitrary class
    or instance dictionaries to give those objects access to stateful
    & ephemeral key material generators.
    """
    self.keyring = Keys(key=key)
    self.akeyring = AsyncKeys(key=key)


def insert_keyrings(self, key=None):
    """
    A generic __init__ function that can be copied into abitrary class
    or instance dictionaries to give those objects access to stateful
    & ephemeral key material generators.
    """
    self.keyring = Keys(key=key)
    self.akeyring = AsyncKeys(key=key)


OneTimePad.__init__ = insert_keyrings


__extras = {
    "AsyncKeys": AsyncKeys,
    "Keys": Keys,
    "__doc__": __doc__,
    "__main_exports__": __all__,
    "__package__": "aiootp",
    "ainsert_keyrings": ainsert_keyrings,
    "insert_keyrings": insert_keyrings,
    "akeys": akeys,
    "keys": keys,
    "asubkeys": asubkeys,
    "subkeys": subkeys,
    "akeypair": akeypair,
    "keypair": keypair,
    "akeypair_ratchets": akeypair_ratchets,
    "keypair_ratchets": keypair_ratchets,
    "atable_key": atable_key,
    "table_key": table_key,
    "atable_key_gen": atable_key_gen,
    "table_key_gen": table_key_gen,
}


keygens = commons.Namespace.make_module("keygens", mapping=__extras)




# This file is part of aiootp, an asynchronous one-time-pad based crypto
# and anonymity library.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2020 Gonzo Investigatory Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2020 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


__all__ = [
    "randoms",
    "arandom_256",
    "random_256",
    "arandom_512",
    "random_512",
    "asafe_symm_keypair",
    "safe_symm_keypair",
    "aseeder",
    "seeder",
    "acsprng",
    "csprng",
]


__doc__ = """
A collection of functions which use & create varying levels of entropy
for end-user cryptographic applications.
"""


import os
import math
from random import choice
from random import uniform
from random import randrange as random_range
from hashlib import sha3_256
from hashlib import sha3_512
from collections import deque
from sympy import prevprime as prev_prime
from sympy import nextprime as next_prime
from sympy import randprime as random_prime
from .asynchs import *
from .asynchs import sleep
from .asynchs import asleep
from .commons import bits
from .commons import primes
from .commons import commons
from .generics import aint
from .generics import astr
from .generics import arange
from .generics import nc_512
from .generics import anc_512
from .generics import sha_256
from .generics import sha_512
from .generics import asha_256
from .generics import asha_512
from .generics import Enumerate
from .generics import Comprende
from .generics import comprehension
from .generics import bytes_to_int
from .generics import abytes_to_int
from .generics import sha_512_hmac
from .generics import asha_512_hmac
from .generics import is_async_iterable


async def acreate_prime(bits=2048):
    """
    Asynchronous wrapper around a ``sympy.randprime`` abstraction which
    locates primes based on a user-defined amount of ``bits``.
    """
    return random_prime(2 ** bits, 2 ** (bits + 1))


def create_prime(bits=2048):
    """
    Synchronous wrapper around a ``sympy.randprime`` abstraction which
    locates primes based on a user-defined amount of ``bits``.
    """
    return random_prime(2 ** bits, 2 ** (bits + 1))


async def arandom_prime(low=None, high=None, **kw):
    """
    Asynchronous wrapper around ``sympy.randprime``.
    """
    return random_prime(low, high, **kw)


async def aprime_table(
    low=None, high=None, step=1, depth=10, depth_error=25
):
    """
    Create a dictionary with `depth` number of random primes per group,
    where range(low, high, step) determines the number of groups, and
    each value and value + 1 in this range represents the minimum and
    maximum number of bits, respectively, each prime in the group will
    be.

    If low or high are passed without the other, then high is assumed to
    be low + 1, and low is assumed to be high - 1, respectively. By
    default, if low and high are not passed, this function produces a
    dictionary with 64 groups of primes from 64 to 128 bits each, with
    groups of size 10 each.

    The depth_error argument is used when attempting to populate a prime
    group with `depth` number of elements, and `depth_error` iterations
    have elapsed which produced a prime that's already in the table.
    This throws an error indicating either that the function may be in
    an infinite loop because the prime group has been completely
    consumed of unique elements, or that the threshhold has been reached
    where the user no longer wants to search the group for unique
    elements.

    Usage Examples:

    table = prime_table(low=128, depth=3)
    print(table[128])
    [
        493533133190526841522032284522322968107,
        388191793051351746548745812175673076809,
        592596411378307857279359758985424061099,
    ]

    table = prime_table(low=128, high=1029, step=128, depth=1)
    print(128 in table)
    >>> True
    print(256 in table)
    >>> True
    print(384 in table)
    >>> True
    print(512 in table)
    >>> True
    print(513 in table)
    >>> False
    """
    if low == None and high == None:
        low = 64
        high = 129
    elif low == None:
        low = high - 1
    elif high == None:
        high = low + 1
    elif low >= high:
        raise ValueError("Lower bound isn't less than the upper bound.")
    table = {}
    for prime_group in range(low, high, step):
        min_bits = 2 ** prime_group
        max_bits = 2 ** (prime_group + 1)
        table.update({prime_group: []})
        for loop in range(depth):
            prime = await arandom_prime(min_bits, max_bits)
            infinite_loop_checker = 0
            while prime in table[prime_group]:
                infinite_loop_checker += 1
                if infinite_loop_checker >= depth_error:
                    problem = "Max recursion depth reached because "
                    cause = "not enough unique primes could be found "
                    cause += f"within the prime group of {prime_group} "
                    cause += f"bits to fill a list of length {depth}."
                    error = RuntimeError(problem + cause)
                    error.value = table
                    raise error
                prime = await arandom_prime(min_bits, max_bits)
            table[prime_group].append(prime)
    return table


def prime_table(low=None, high=None, step=1, depth=10, depth_error=25):
    """
    Create a dictionary with `depth` number of random primes per group,
    where range(low, high, step) determines the number of groups, and
    each value and value + 1 in this range represents the minimum and
    maximum number of bits, respectively, each prime in the group will
    be.

    If low or high are passed without the other, then high is assumed to
    be low + 1, and low is assumed to be high - 1, respectively. By
    default, if low and high are not passed, this function produces a
    dictionary with 64 groups of primes from 64 to 128 bits each, with
    groups of size 10 each.

    The depth_error argument is used when attempting to populate a prime
    group with `depth` number of elements, and `depth_error` iterations
    have elapsed which produced a prime that's already in the table.
    This throws an error indicating either that the function may be in
    an infinite loop because the prime group has been completely
    consumed of unique elements, or that the threshhold has been reached
    where the user no longer wants to search the group for unique
    elements.

    Usage Examples:

    table = prime_table(low=128, depth=3)
    print(table[128])
    [
        493533133190526841522032284522322968107,
        388191793051351746548745812175673076809,
        592596411378307857279359758985424061099,
    ]

    table = prime_table(low=128, high=1029, step=128, depth=1)
    print(128 in table)
    >>> True
    print(256 in table)
    >>> True
    print(384 in table)
    >>> True
    print(512 in table)
    >>> True
    print(513 in table)
    >>> False
    """
    if low == None and high == None:
        low = 64
        high = 129
    elif low == None:
        low = high - 1
    elif high == None:
        high = low + 1
    elif low >= high:
        raise ValueError("Lower bound isn't less than the upper bound.")
    table = {}
    for prime_group in range(low, high, step):
        min_bits = 2 ** prime_group
        max_bits = 2 ** (prime_group + 1)
        table.update({prime_group: []})
        for loop in range(depth):
            prime = random_prime(min_bits, max_bits)
            infinite_loop_checker = 0
            while prime in table[prime_group]:
                infinite_loop_checker += 1
                if infinite_loop_checker >= depth_error:
                    problem = "Max recursion depth reached because "
                    cause = "not enough unique primes could be found "
                    cause += f"within the prime group of {prime_group} "
                    cause += f"bits to fill a list of length {depth}."
                    error = RuntimeError(problem + cause)
                    error.value = table
                    raise error
                prime = random_prime(min_bits, max_bits)
            table[prime_group].append(prime)
    return table


async def auniform(*a, **kw):
    """
    Asynchronous version of the standard library's ``random.uniform``.
    """
    return uniform(*a, **kw)


async def achoice(iterable):
    """
    Asynchronous version of the standard library's ``random.choice``.
    """
    return choice(iterable)


async def arandom_range(*a, **kw):
    """
    Asynchronous version of the standard library's ``random.randrange``.
    """
    return random_range(*a, **kw)


@comprehension()
async def arandom_range_gen(low=1, high=10):
    """
    A generator which produces values from ``random.randrange`` from
    ``low`` to ``high``.
    """
    while True:
        yield random_range(low, high)


@comprehension()
def random_range_gen(low=1, high=10):
    """
    A generator which produces values from ``random.randrange`` from
    ``low`` to ``high``.
    """
    while True:
        yield random_range(low, high)


async def arandom_sleep(span=2):
    """
    Asynchronously sleeps for a psuedo-random portion of ``span`` time.
    """
    return await asleep(span * await auniform(0, 1))


def random_sleep(span=2):
    """
    Synchronously sleeps for a psuedo-random portion of ``span`` time.
    """
    return sleep(span * uniform(0, 1))


async def aurandom_hash(size):
    """
    Returns ``size`` bytes of ``os.urandom`` entropy as a ``sha3_512``
    hash.
    """
    return sha3_512(await aurandom(size)).hexdigest()


def urandom_hash(size):
    """
    Returns ``size`` bytes of ``os.urandom`` entropy as a ``sha3_512``
    hash.
    """
    return sha3_512(urandom(size)).hexdigest()


async def aurandom_number(size):
    """
    Returns ``size`` bytes of ``os.urandom`` entropy as an integer.
    """
    return await abytes_to_int(await aurandom(size))


def urandom_number(size):
    """
    Returns ``size`` bytes of ``os.urandom`` entropy as an integer.
    """
    return bytes_to_int(urandom(size))


async def aurandom(size):
    """
    Returns ``size`` bytes of ``os.urandom`` entropy.
    """
    return os.urandom(size)


def urandom(size):
    """
    Returns ``size`` bytes of ``os.urandom`` entropy.
    """
    return os.urandom(size)


@comprehension()
async def asalted_multiply(mod=primes[258][-1], offset=None):
    """
    Allows for non-communitive multiplication. This assists pseudo-random
    number generators in turning combinations of low entropy number
    sources into permutations. This greatly increases the amount of
    knowledge an attacker must have to perform pre-image or known-
    plaintext attacks on the hashes of those permutations.

    ``mod``:    Should be a prime of bit-size matching the key space
        needs of its use case. Defaults to the last 258-bit prime
        number.
    ``offset``: A number which is used to salt the calculations. It
        should be a pseudo-random number >128-bits larger than ``mod``.

    Usage Example:

    numbers = list(range(10))
    multiply = await asalted_multiply().aprime()
    randomized_number = await multiply(numbers)
    """
    if offset == None:
        offset = await arandom_range(
            bits[128] * mod + 1, (bits[128] + 1) * mod
        )
    mix = offset * next_prime(math.log2(offset)) % mod
    start = seed = int(await asha_256(mix, offset), 16)
    numbers = (offset * seed,)
    while True:
        mix ^= abs(sum((seed, offset, *numbers)))
        mix %= mod
        seed ^= mix
        offset ^= seed
        (*numbers,) = yield start ^ seed
        start %= mod
        for number in numbers:
            mix += offset
            start *= number ^ mix
            await switch()


@comprehension()
def salted_multiply(mod=primes[258][-1], offset=None):
    """
    Allows for non-communitive multiplication. This assists pseudo-random
    number generators in turning combinations of low entropy number
    sources into permutations. This greatly increases the amount of
    knowledge an attacker must have to perform pre-image or known-
    plaintext attacks on the hashes of those permutations.

    ``mod``:    Should be a prime of bit-size matching the key space
        needs of its use case. Defaults to the last 258-bit prime
        number.
    ``offset``: A number which is used to salt the calculations. It
        should be a pseudo-random number >128-bits larger than ``mod``.

    Usage Example:

    numbers = list(range(10))
    multiply = salted_multiply().prime()
    randomized_number = multiply(numbers)
    """
    if offset == None:
        offset = random_range(bits[128] * mod + 1, (bits[128] + 1) * mod)
    mix = offset * next_prime(math.log2(offset)) % mod
    start = seed = int(sha_256(mix, offset), 16)
    numbers = (offset * seed,)
    while True:
        mix ^= abs(sum((seed, offset, *numbers)))
        mix %= mod
        seed ^= mix
        offset ^= seed
        (*numbers,) = yield start ^ seed
        start %= mod
        for number in numbers:
            mix += offset
            start *= number ^ mix


#  initializing weakly entropic coroutines
mod = primes[512][0]

_asalt_multiply = run(asalted_multiply(mod).aprime())
_salt_multiply = salted_multiply(mod).prime()

_initial_entropy = deque(
    [urandom_number(128), urandom_number(128)], maxlen=2
)
del mod


async def asalt(*, _entropy=_initial_entropy):
    """
    Returns a low-grade entropy number from cached & ratcheted system
    entropy.
    """
    _entropy.appendleft(int(await asha_512(_entropy), 16))
    return await _asalt_multiply(_entropy)


def salt(*, _entropy=_initial_entropy):
    """
    Returns a low-grade entropy number from cached & ratcheted system
    entropy.
    """
    _entropy.appendleft(int(sha_512(_entropy), 16))
    return _salt_multiply(_entropy)


async def arandom_256(entropy=salt(), runs=26, refresh=False):
    """
    Returns a 256-bit hash produced by ``random_number_generator``.
    Users can pass ``entropy`` into the function. ``runs`` determines
    how many internal iterations ``random_number_generator`` will crank
    on when deriving a new key. ``refresh`` is ``False`` by default,
    since generating new entropy can be quite slow -- it can be toggled
    to ``True`` to generate new entropy.
    """
    return await asha_256(
        await arandom_number_generator(
            entropy=entropy, runs=runs, refresh=refresh
        ),
        await aurandom(32),
    )


def random_256(entropy=salt(), runs=26, refresh=False):
    """
    Returns a 256-bit hash produced by ``random_number_generator``.
    Users can pass ``entropy`` into the function. ``runs`` determines
    how many internal iterations ``random_number_generator`` will crank
    on when deriving a new key. ``refresh`` is ``False`` by default,
    since generating new entropy can be quite slow -- it can be toggled
    to ``True`` to generate new entropy.
    """
    return sha_256(
        random_number_generator(
            entropy=entropy, runs=runs, refresh=refresh
        ),
        urandom(32),
    )


async def arandom_512(entropy=salt(), runs=26, refresh=False):
    """
    Returns a 512-bit hash produced by ``random_number_generator``.
    Users can pass ``entropy`` into the function. ``runs`` determines
    how many internal iterations ``random_number_generator`` will crank
    on when deriving a new key. ``refresh`` is ``False`` by default,
    since generating new entropy can be quite slow -- it can be toggled
    to ``True`` to generate new entropy.
    """
    return await asha_512(
        await arandom_number_generator(
            entropy=entropy, runs=runs, refresh=refresh
        ),
        await aurandom(64),
    )


def random_512(entropy=salt(), runs=26, refresh=False):
    """
    Returns a 512-bit hash produced by ``random_number_generator``.
    Users can pass ``entropy`` into the function. ``runs`` determines
    how many internal iterations ``random_number_generator`` will crank
    on when deriving a new key. ``refresh`` is ``False`` by default,
    since generating new entropy can be quite slow -- it can be toggled
    to ``True`` to generate new entropy.
    """
    return sha_512(
        random_number_generator(
            entropy=entropy, runs=runs, refresh=refresh
        ),
        urandom(64),
    )


async def arandom_number_generator(
    entropy=salt(),
    runs=26,
    refresh=False,
    *,
    _completed=deque([], maxlen=256),
):
    """
    We propose several methods for producing cryptographically secure
    pseudo-random numbers, and implement them in this function.

    1. Greatly increase the space an attacker must search by using
    pseudo-random seeds that are very large numbers (> 256-bits and
    < 2048-bits) as the range arguments for random.randrange.

    2. Recursively use cryptographic hashing on the results of our
    pseudo-random number generators to unlink their mathematical
    associations from their seeds.

    3. Frustrate attempts to learn the internal state of the PRNGs by
    making separate seeds at each import-time & function call, as well
    as, incorporating forward-secure key ratcheting algorithms within
    several of this package's pseudo-random number generators. They work
    together, with proper coordination, to produce randomness in a way
    that alone they wouldn't. Using a combination of PRNGs also spreads
    out the impact of an attacker predicting CSPRNG output by requiring
    all of the PRNGs be broken at import time, or for the user machine's
    memory to be maliciously analyzed, to predict outputs from the
    CSPRNGs.

    4. Use three randrange function abstractions where for each function
    call: one produces 2 new small pseudo-random seeds from a pair of
    smaller seeds (256-bits - 512-bits); the second produces 2 new
    large pseudo-random seeds from the pair of larger seeds (1536-bits -
    2048-bits); the third to produce a large pseudo-random number from
    ``random.randrange`` with the lower bound being the two small seeds
    multiplied together and salted, and the upper bound being the two
    large seeds being multiplied together and salted.

    5. Further frustrate an attacker by necessitating that they perform
    an order prediction attack on the results of the CSPRNG's component
    PRNGs. We do this by using asynchrony entirely throughout the CSPRNG
    with random sleeps.

    6. Iteratively use a salted modular multiplication on the CSPRNG's
    component PRNG states, then hashing & caching the results.

    7. Use the powerful sha3_512 hashing algorithm for all hashing.

    8. Iterate and interleave all these methods enough times such that if
    we assume each time the CSPRNG produces a 512-bit internal state,
    that only 1-bit of entropy is produced. Then, by initializing up to
    a cache of 256 ratcheting states, then the ``random_number_generator``
    algorithm here would have at least 256-bits of entropy.

    **** **** **** **** **** **** **** **** **** **** **** **** ****
    Our implementation analysis is NOT vigorous or conclusive, though
    soley based on the constraints stated above, our assumptions of
    randomness should be a reasonable estimate.
    **** **** **** **** **** **** **** **** **** **** **** **** ****
    """
    if refresh or not _completed:

        async def start_generator(runs, tasks=deque()):
            for _ in range(runs):
                await switch()
                tasks.appendleft(new_task(modular_exponentiation()))
                for _ in range(10):
                    tasks.appendleft(new_task(hash_cache()))
            await gather(*tasks)

        async def hash_cache(seed=await aurandom(64)):
            await arandom_sleep(0.001)
            _completed.appendleft(await asha_512(_completed, entropy, seed))

        async def modular_exponentiation(seed=await asalt()):
            multiples = deque()
            for _ in range(3):
                await arandom_sleep(0.001)
                multiples.appendleft(
                    new_task(create_unique_multiple(seed))
                )
            result = await big_modulation(
                seed, *[await calculation for calculation in multiples]
            )
            _completed.appendleft(await asha_512(result, seed))

        async def create_unique_multiple(seed):
            return await big_multiply(
                seed,
                await aunique_integer(),
                await aunique_integer(),
                await aunique_integer(),
            )

        async def big_modulation(seed, multiple_0, multiple_1, multiple_2):
            return await big_multiply(
                seed,
                multiple_0,
                multiple_1,
                multiple_2,
                await aunique_big_int(),
            ) % await achoice(primes[4096])

        async def big_multiply(*args):
            return await _asalt_multiply(args)

        await agenerate_unique_range_bounds()
        await start_generator(runs)
        await agenerate_unique_range_bounds()
    else:
        _completed.appendleft(
            await asha_512_hmac((_completed, salt()), entropy)
        )

    return await asha_512_hmac((_completed, salt()), entropy)


def random_number_generator(
    entropy=salt(),
    runs=26,
    refresh=False,
    *,
    _completed=deque([], maxlen=256),
):
    """
    We propose several methods for producing cryptographically secure
    pseudo-random numbers, and implement them in this function.

    1. Greatly increase the space an attacker must search by using
    pseudo-random seeds that are very large numbers (> 256-bits and
    < 2048-bits) as the range arguments for random.randrange.

    2. Recursively use cryptographic hashing on the results of our
    pseudo-random number generators to unlink their mathematical
    associations from their seeds.

    3. Frustrate attempts to learn the internal state of the PRNGs by
    making separate seeds at each import-time & function call, as well
    as, incorporating forward-secure key ratcheting algorithms within
    several of this package's pseudo-random number generators. They work
    together, with proper coordination, to produce randomness in a way
    that alone they wouldn't. Using a combination of PRNGs also spreads
    out the impact of an attacker predicting CSPRNG output by requiring
    all of the PRNGs be broken at import time, or for the user machine's
    memory to be maliciously analyzed, to predict outputs from the
    CSPRNGs.

    4. Use three randrange function abstractions where for each function
    call: one produces 2 new small pseudo-random seeds from a pair of
    smaller seeds (256-bits - 512-bits); the second produces 2 new
    large pseudo-random seeds from the pair of larger seeds (1536-bits -
    2048-bits); the third to produce a large pseudo-random number from
    ``random.randrange`` with the lower bound being the two small seeds
    multiplied together and salted, and the upper bound being the two
    large seeds being multiplied together and salted.

    5. Further frustrate an attacker by necessitating that they perform
    an order prediction attack on the results of the CSPRNG's component
    PRNGs. We do this by using asynchrony entirely throughout the CSPRNG
    with random sleeps.

    6. Iteratively use a salted modular multiplication on the CSPRNG's
    component PRNG states, then hashing & caching the results.

    7. Use the powerful sha3_512 hashing algorithm for all hashing.

    8. Iterate and interleave all these methods enough times such that,
    we assume each time the CSPRNG produces a 512-bit internal state,
    that only 1-bit of entropy is produced. Then, by initializing up to
    a cache of 256 ratcheting states then the ``random_number_generator``
    algorithm here would have at least 256-bits of entropy.

    **** **** **** **** **** **** **** **** **** **** **** **** ****
    Our implementation analysis is NOT vigorous or conclusive, though
    soley based on the constraints stated above, our assumptions of
    randomness should be a reasonable estimate.
    **** **** **** **** **** **** **** **** **** **** **** **** ****
    """
    if refresh or not _completed:

        async def start_generator(runs, tasks=deque()):
            for _ in range(runs):
                await switch()
                tasks.appendleft(new_task(modular_exponentiation()))
                for _ in range(10):
                    tasks.appendleft(new_task(hash_cache()))
            await gather(*tasks)

        async def hash_cache(seed=urandom(64)):
            await arandom_sleep(0.001)
            _completed.appendleft(await asha_512(_completed, entropy, seed))

        async def modular_exponentiation(seed=salt()):
            multiples = deque()
            for _ in range(3):
                await arandom_sleep(0.001)
                multiples.appendleft(
                    new_task(create_unique_multiple(seed))
                )
            result = await big_modulation(
                seed, *[await calculation for calculation in multiples]
            )
            _completed.appendleft(await asha_512(result, seed))

        async def create_unique_multiple(seed):
            return await big_multiply(
                seed,
                await aunique_integer(),
                await aunique_integer(),
                await aunique_integer(),
            )

        async def big_modulation(seed, multiple_0, multiple_1, multiple_2):
            return await big_multiply(
                seed,
                multiple_0,
                multiple_1,
                multiple_2,
                await aunique_big_int(),
            ) % await achoice(primes[4096])

        async def big_multiply(*args):
            return await _asalt_multiply(args)

        generate_unique_range_bounds()
        run(start_generator(runs))
        generate_unique_range_bounds()
    else:
        _completed.appendleft(sha_512_hmac((_completed, salt()), entropy))

    return sha_512_hmac((_completed, salt()), entropy)


async def aunique_integer():
    """
    Returns an ``int(hex_hash, 16)`` value of a unique hexidecimal hash.
    """
    return int(await aunique_hash(), 16)


def unique_integer():
    """
    Returns an ``int(hex_hash, 16)`` value of a unique hexidecimal hash.
    """
    return int(unique_hash(), 16)


async def aunique_hash():
    """
    Returns a ``hashlib.sha3_512`` string hash of an integer which is
    greater than a 512-bit number by many orders of magnitude.
    """
    return sha_512(await aunique_big_int())


def unique_hash():
    """
    Returns a ``hashlib.sha3_512`` string hash of an integer which is
    greater than a 512-bit number by many orders of magnitude.
    """
    return sha_512(unique_big_int())


async def aunique_big_int():
    """
    Uses unique lower & upper bound integers to feed into the standard
    library's ``randrange`` function & returns the result.
    """
    return await arandom_range(
        *(await gather(aunique_lower_bound(), aunique_upper_bound()))
    )


def unique_big_int():
    """
    Uses unique lower & upper bound integers to feed into the standard
    library's ``randrange`` function & returns the result.
    """
    return random_range(unique_lower_bound(), unique_upper_bound())


async def aunique_lower_bound():
    """
    Returns a unique number where 2**1536 < number < 2**2048 from a pair
    of global, semi-constant 256-bit - 512-bit seeds.
    """
    global SMALL_UPPER_BOUND
    global SMALL_LOWER_BOUND
    number_0 = arandom_range(SMALL_LOWER_BOUND, SMALL_UPPER_BOUND)
    number_1 = arandom_range(SMALL_LOWER_BOUND, SMALL_UPPER_BOUND)
    return await _asalt_multiply(await gather(number_0, number_1))


def unique_lower_bound():
    """
    Returns a unique number where 2**1536 < number < 2**2048 from a pair
    of global, semi-constant 256-bit - 512-bit seeds.
    """
    global SMALL_UPPER_BOUND
    global SMALL_LOWER_BOUND
    return _salt_multiply(
        [
            random_range(SMALL_LOWER_BOUND, SMALL_UPPER_BOUND),
            random_range(SMALL_LOWER_BOUND, SMALL_UPPER_BOUND),
        ]
    )


async def aunique_upper_bound():
    """
    Returns a unique number where 2**4096 < number < 2**4608 from a pair
    of global, semi-constant 1536-bit - 2048-bit seeds.
    """
    global BIG_UPPER_BOUND
    global BIG_LOWER_BOUND
    number_0 = arandom_range(BIG_LOWER_BOUND, BIG_UPPER_BOUND)
    number_1 = arandom_range(BIG_LOWER_BOUND, BIG_UPPER_BOUND)
    return await _asalt_multiply(await gather(number_0, number_1))


def unique_upper_bound():
    """
    Returns a unique number where 2**4096 < number < 2**4608 from a pair
    of global, semi-constant 1536-bit - 2048-bit seeds.
    """
    global BIG_UPPER_BOUND
    global BIG_LOWER_BOUND
    return _salt_multiply(
        [
            random_range(BIG_LOWER_BOUND, BIG_UPPER_BOUND),
            random_range(BIG_LOWER_BOUND, BIG_UPPER_BOUND),
        ]
    )


async def agenerate_unique_range_bounds():
    """
    Generates two pairs of unique global, semi-constant seeds which
    feed uniqueness into ``random.randrange``, with the consideration
    that guessing its output is aided by knowing what its inputs were.
    Making its inputs unknown should then help keep its outputs unknown.
    """
    await gather(
        agenerate_small_range_bounds(), agenerate_big_range_bounds()
    )


def generate_unique_range_bounds():
    """
    Generates two pairs of unique global, semi-constant seeds which
    feed uniqueness into ``random.randrange``, with the consideration
    that guessing its output is aided by knowing what its inputs were.
    Making its inputs unknown should then help keep its outputs unknown.
    """
    generate_small_range_bounds()
    generate_big_range_bounds()


async def agenerate_small_range_bounds():
    """
    Generates a pair of unique global, semi-constant seeds which feed
    uniqueness into the lower bound of ``random.randrange``, with the
    consideration that guessing its output is aided by knowing what its
    inputs were. Making its inputs unknown should then help keep its
    outputs unknown.
    """
    global SMALL_UPPER_BOUND
    global SMALL_LOWER_BOUND
    SMALL_UPPER_BOUND = await atemplate_unique_number(bits[512])
    SMALL_LOWER_BOUND = await atemplate_unique_number(bits[256])


def generate_small_range_bounds():
    """
    Generates a pair of unique global, semi-constant seeds which feed
    uniqueness into the lower bound of ``random.randrange``, with the
    consideration that guessing its output is aided by knowing what its
    inputs were. Making its inputs unknown should then help keep its
    outputs unknown.
    """
    global SMALL_UPPER_BOUND
    global SMALL_LOWER_BOUND
    SMALL_UPPER_BOUND = template_unique_number(bits[512])
    SMALL_LOWER_BOUND = template_unique_number(bits[256])


async def agenerate_big_range_bounds():
    """
    Generates a pair of unique global, semi-constant seeds which feed
    uniqueness into the upper bound of ``random.randrange``, with the
    consideration that guessing its output is aided by knowing what its
    inputs were. Making its inputs unknown should then help keep its
    outputs unknown.
    """
    global BIG_UPPER_BOUND
    global BIG_LOWER_BOUND
    BIG_UPPER_BOUND = await atemplate_unique_number(bits[2048])
    BIG_LOWER_BOUND = await atemplate_unique_number(bits[1536])


def generate_big_range_bounds():
    """
    Generates a pair of unique global, semi-constant seeds which feed
    uniqueness into the upper bound of ``random.randrange``, with the
    consideration that guessing its output is aided by knowing what its
    inputs were. Making its inputs unknown should then help keep its
    outputs unknown.
    """
    global BIG_UPPER_BOUND
    global BIG_LOWER_BOUND
    BIG_UPPER_BOUND = template_unique_number(bits[2048])
    BIG_LOWER_BOUND = template_unique_number(bits[1536])


async def atemplate_unique_number(number):
    """
    A pseudo-random number generator helper function. An alternative
    method of constructing unique numbers. The length of the number
    argument will be the same as the length of the number that's
    returned. This function is used to produce pseudo-random numbers for
    the ranges passed to random.randrange. We assume that, alone, the
    output of random.randrange can be determined by an attacker using
    several known or unknown attack vectors. So this function is to be
    used in conjunction with other, vigorous methods of producing
    cryptographically secure pseudo-random numbers.
    """
    seed = await asalt()
    number = await aint(number)  # throw if not a number
    while seed < number:
        seed *= await asalt()
    return await aint(str(seed)[: len(str(number))])


def template_unique_number(number):
    """
    A pseudo-random number generator helper function. An alternative
    method of constructing unique numbers. The length of the number
    argument will be the same as the length of the number that's
    returned. This function is used to produce pseudo-random numbers for
    the ranges passed to random.randrange. We assume that, alone, the
    output of random.randrange can be determined by an attacker using
    several known or unknown attack vectors. So this function is to be
    used in conjunction with other, vigorous methods of producing
    cryptographically secure pseudo-random numbers.
    """
    seed = salt()
    number = int(number)  # throw if not a number
    while seed < number:
        seed *= salt()
    return int(str(seed)[: len(str(number))])


async def asafe_symm_keypair(entropy=salt(), refresh=False, runs=26):
    """
    Returns two ``bytes`` type, 512-bit hexidecimal keys. This function
    updates the package's static random seeds before & after deriving
    the returned key pair.
    """
    global global_seed
    global global_seed_key

    await agenerate_unique_range_bounds()
    seed_key = sha_512(
        global_seed,
        global_seed_key,
        await arandom_512(refresh=refresh, runs=runs),
    ).encode()
    seed = sha_512_hmac((global_seed, seed_key), key=entropy).encode()
    global_seed_key = sha_512_hmac(seed, global_seed_key)
    global_seed = sha_512_hmac(global_seed, global_seed_key)
    await agenerate_unique_range_bounds()
    return seed, seed_key


def safe_symm_keypair(entropy=salt(), refresh=False, runs=26):
    """
    Returns two ``bytes`` type, 512-bit hexidecimal keys. This function
    updates the package's static random seeds before & after deriving
    the returned key pair.
    """
    global global_seed
    global global_seed_key

    generate_unique_range_bounds()
    seed_key = sha_512(
        global_seed, global_seed_key, random_512(refresh=refresh, runs=runs)
    ).encode()
    seed = sha_512_hmac((global_seed, seed_key), key=entropy).encode()
    global_seed_key = sha_512_hmac(seed, global_seed_key)
    global_seed = sha_512_hmac(global_seed, global_seed_key)
    generate_unique_range_bounds()
    return seed, seed_key


@comprehension()
async def aseeder(entropy=salt(), refresh=False, runs=26):
    """
    A fast random number generator that supports adding entropy during
    async iteration. It's based on the randomness produced from combining
    a key ratchet algorithm, os psuedo randomness, and this module's
    cryptographically secure pseudo-random number generator (csprng).

    Usage examples:

    # In async for loops ->
    async for seed in aseeder():
        # do something with ``seed``, a random hash

    # By calling anext ->
    from aiootp import anext
    from aioitertools import iter as aiter

    csprng = aseeder()
    seed = await next(csprng)

    # By sending in entropy ->
    entropy = "any object whose str() representation has randomness"
    csprng = aseeder(entropy)  # entropy can be added here
    await csprng(None)
    seed = await csprng(entropy)  # &/or entropy can be added here
    """
    seed, seed_key = await asafe_symm_keypair(entropy, refresh)
    kdf = sha3_512(seed + seed_key)
    while True:
        entropy = (await astr(entropy)).encode()
        kdf.update(seed + seed_key + entropy)
        seed_key = kdf.digest()
        kdf.update(seed + seed_key + entropy)
        seed = kdf.digest()
        kdf.update(seed + seed_key + entropy)
        entropy = yield kdf.hexdigest()


@comprehension()
def seeder(entropy=salt(), refresh=False, runs=26):
    """
    A fast random number generator that supports adding entropy during
    iteration. It's based on the randomness produced from combining a
    key ratchet algorithm, os psuedo randomness, and this module's
    cryptographically secure pseudo-random number generator (csprng).

    Usage examples:

    # In for loops ->
    for seed in seeder():
        # do something with ``seed``, a random hash

    # By calling next ->
    csprng = seeder()
    seed = next(csprng)

    # By sending in entropy ->
    entropy = "any object whose str() representation has randomness"
    csprng = seeder(entropy)  # entropy can be added here
    csprng(None)
    seed = csprng(entropy)  # &/or entropy can be added here
    """
    seed, seed_key = safe_symm_keypair(entropy, refresh, runs)
    kdf = sha3_512(seed + seed_key)
    while True:
        entropy = str(entropy).encode()
        kdf.update(seed + seed_key + entropy)
        seed_key = kdf.digest()
        kdf.update(seed + seed_key + entropy)
        seed = kdf.digest()
        kdf.update(seed + seed_key + entropy)
        entropy = yield kdf.hexdigest()


@comprehension()
async def anon0_digit_stream(key=salt(), stream_key=""):
    """
    Creates a deterministic stream of non-zero digits from a key.
    """
    seed = await asha_512(key)
    while True:
        stream_key = await aint(await anc_512(seed, key, stream_key), 16)
        for char in (await astr(stream_key))[4:].replace("0", ""):
            yield await aint(char)


@comprehension()
def non0_digit_stream(key=salt(), stream_key=""):
    """
    Creates a deterministic stream of non-zero digits from a key.
    """
    seed = sha_512(key)
    while True:
        stream_key = int(nc_512(seed, key, stream_key), 16)
        for char in (str(stream_key)[4:]).replace("0", ""):
            yield int(char)


@comprehension()
async def adigit_stream(key=salt(), stream_key=""):
    """
    Creates a deterministic stream of digits from a key.
    """
    seed = await asha_512(key)
    while True:
        stream_key = await aint(await anc_512(seed, key, stream_key), 16)
        for char in (await astr(stream_key))[4:]:
            yield await aint(char)


@comprehension()
def digit_stream(key=salt(), stream_key=""):
    """
    Creates a deterministic stream of digits from a key.
    """
    seed = sha_512(key)
    while True:
        stream_key = int(nc_512(seed, key, stream_key), 16)
        for char in str(stream_key)[4:]:
            yield int(char)


@comprehension()
async def aleaf_walk(
    stamp="", left=frozenset("02468ace"), right=frozenset("13579bdf")
):
    """
    Walks a binary choice tree according to membership tests for each
    item in ``stamp`` on the two iterables ``left `` & ``right``. Yields
    the current step number and a boolean value, ``True`` if the current
    item of ``stamp`` is in ``left``, else ``False`` if it's in ``right``.
    """
    async for num, turn in Enumerate(stamp, start=1):
        if turn in left:
            yield num, True
        elif turn in right:
            yield num, False


@comprehension()
def leaf_walk(
    stamp="", left=frozenset("02468ace"), right=frozenset("13579bdf")
):
    """
    Walks a binary choice tree according to membership tests for each
    item in ``stamp`` on the two iterables ``left `` & ``right``. Yields
    the current step number and a boolean value, ``True`` if the current
    item of ``stamp`` is in ``left``, else ``False`` if it's in ``right``.
    """
    for num, turn in enumerate(stamp, start=1):
        if turn in left:
            yield num, True
        elif turn in right:
            yield num, False


async def aleaf(
    stamp="",
    mod=bits[256],
    left=frozenset("02468ace"),
    right=frozenset("13579bdf"),
    total=primes[256][0],
):
    """
    Returns a unique name of type ``int`` based on the permutations of
    ``stamp``'s elements' membership in ``left`` and ``right``.
    """
    prime = primes[16][0]
    async for num, turn in aleaf_walk(stamp, left, right):
        if turn:
            total += num
        else:
            total *= prime
    return total % mod


def leaf(
    stamp="",
    max_binary_choices=bits[256],
    left=frozenset("02468ace"),
    right=frozenset("13579bdf"),
    prime_strength_bits=primes[256][0],
):
    """
    Returns a unique name of type ``int`` based on the permutations of
    ``stamp``'s elements' membership in ``left`` and ``right``.
    """
    total = prime_strength_bits
    prime_multiple = primes[16][0]
    for step, turn in leaf_walk(stamp, left, right):
        if turn:
            total += step
        else:
            total *= prime_multiple
    return total % max_binary_choices


@comprehension()
async def amake_uuid(length=16, salt=None):
    """
    Creates a deterministic, unique user id from a ``salt`` & a ``stamp``
    sent into coroutine.
    """
    stamp = None
    multiple = (length // 512) + 1
    salt = salt if salt != None else await acsprng()
    async with Comprende().arelay(result=salt):
        while True:
            uuid = ""
            async for growth in arange(multiple):
                uuid += sha_512(uuid, stamp, salt)
            stamp = yield uuid[-1:-(1 + length):-1]


@comprehension()
def make_uuid(length=16, salt=None):
    """
    Creates a deterministic, unique user id from a ``salt`` & a ``stamp``
    sent into coroutine.
    """
    stamp = None
    multiple = (length // 512) + 1
    salt = salt if salt != None else csprng()
    with Comprende().relay(result=salt):
        while True:
            uuid = ""
            for growth in range(multiple):
                uuid += sha_512(uuid, stamp, salt)
            stamp = yield uuid[-1:-(1 + length):-1]


try:
    # Initalize package entropy pool & cryptographically secure pseudo-
    # random number generators.
    global_seed_key = random_512(entropy=salt())
    global_seed = random_512(entropy=global_seed_key)
    csprng = seeder(global_seed).send
    acsprng = aseeder(csprng()).asend
    global_seed_key = run(acsprng())
    global_seed = run(acsprng(csprng(csprng())))
except RuntimeError as error:
    problem = "package's random seed initialization failed likely because "
    location = f"{__name__} from {__package__} "
    reason = f"was imported from within an async event loop."
    failure = RuntimeError(problem + location + reason)
    raise failure from error


__extras = {
    "__doc__": __doc__,
    "__main_exports__": __all__,
    "__package__": "aiootp",
    "achoice": achoice,
    "acsprng": acsprng,
    "adigit_stream": adigit_stream,
    "agenerate_big_range_bounds": agenerate_big_range_bounds,
    "agenerate_small_range_bounds": agenerate_small_range_bounds,
    "agenerate_unique_range_bounds": agenerate_unique_range_bounds,
    "aleaf": aleaf,
    "aleaf_walk": aleaf_walk,
    "amake_uuid": amake_uuid,
    "anon0_digit_stream": anon0_digit_stream,
    "aprime_table": aprime_table,
    "arandom_256": arandom_256,
    "arandom_512": arandom_512,
    "arandom_number_generator": arandom_number_generator,
    "arandom_prime": arandom_prime,
    "arandom_range": arandom_range,
    "arandom_range_gen": arandom_range_gen,
    "arandom_sleep": arandom_sleep,
    "asafe_symm_keypair": asafe_symm_keypair,
    "asalt": asalt,
    "asalted_multiply": asalted_multiply,
    "aseeder": aseeder,
    "atemplate_unique_number": atemplate_unique_number,
    "auniform": auniform,
    "aunique_big_int": aunique_big_int,
    "aunique_hash": aunique_hash,
    "aunique_integer": aunique_integer,
    "aunique_lower_bound": aunique_lower_bound,
    "aunique_upper_bound": aunique_upper_bound,
    "aurandom": aurandom,
    "aurandom_hash": aurandom_hash,
    "aurandom_number": aurandom_number,
    "choice": choice,
    "csprng": csprng,
    "digit_stream": digit_stream,
    "generate_big_range_bounds": generate_big_range_bounds,
    "generate_small_range_bounds": generate_small_range_bounds,
    "generate_unique_range_bounds": generate_unique_range_bounds,
    "leaf": leaf,
    "leaf_walk": leaf_walk,
    "make_uuid": make_uuid,
    "next_prime": next_prime,
    "non0_digit_stream": non0_digit_stream,
    "prev_prime": prev_prime,
    "prime_table": prime_table,
    "random_256": random_256,
    "random_512": random_512,
    "random_number_generator": random_number_generator,
    "random_prime": random_prime,
    "random_range": random_range,
    "random_range_gen": random_range_gen,
    "random_sleep": random_sleep,
    "safe_symm_keypair": safe_symm_keypair,
    "salt": salt,
    "salted_multiply": salted_multiply,
    "seeder": seeder,
    "template_unique_number": template_unique_number,
    "uniform": uniform,
    "unique_big_int": unique_big_int,
    "unique_hash": unique_hash,
    "unique_integer": unique_integer,
    "unique_lower_bound": unique_lower_bound,
    "unique_upper_bound": unique_upper_bound,
    "urandom": urandom,
    "urandom_hash": urandom_hash,
    "urandom_number": urandom_number,
}


randoms = commons.Namespace.make_module("randoms", mapping=__extras)




# This file is part of aiootp, an asynchronous one-time-pad based crypto
# and anonymity library.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2020 Gonzo Investigatory Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2020 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


__all__ = [
    "asynchs",
    "run",
    "gather",
    "switch",
    "new_task",
    "new_future",
]


__doc__ = """
A collection of asyncio & uvloop references to simplify the standard
usage of those libraries in this package. This module can be used to
replace the default event loop policy, for instance, to run uvloop. The
default asyncio loop is still available in ``default_loop``.
"""


# import uvloop
import asyncio
from time import time
from time import sleep
from .commons import commons


default_loop = asyncio.get_event_loop()


# # Optionally turn on ``uvloop``
# asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())


# # Optionally turn off ``uvloop``
# asyncio.set_event_loop(default_loop)


# # Optionally, set asyncio's debug mode on
# loop().set_debug(True)


def loop(*a, _default=asyncio.get_event_loop, **kw):
    """
    Proxy's access to ``asyncio.get_event_loop()``.
    """
    return _default(*a, **kw)


def serve(*a, **kw):
    """
    Proxy's access to ``asyncio.get_event_loop().run_forever()``.
    """
    return loop().run_forever(*a, **kw)


def run(coro, **kw):
    """
    Proxy's access to ``asyncio.get_event_loop().run_until_complete()``.
    """
    return loop().run_until_complete(coro, **kw)


def new_task(coro, **kw):
    """
    Proxy's access to ``asyncio.get_event_loop().create_taksk()``.
    """
    return loop().create_task(coro, **kw)


def new_future(coro, **kw):
    """
    Proxy's access to ``asyncio.ensure_future()``.
    """
    return asyncio.ensure_future(coro, **kw)


def gather(*coros, **kw):
    """
    Proxy's access to ``asyncio.gather()``.
    """
    return asyncio.gather(*coros, **kw)


async def asleep(delay=0.0, **kw):
    """
    Proxy's access to ``asyncio.sleep()``.
    """
    return await asyncio.sleep(delay, **kw)


def stop():
    """
    Stops the currently running event loop.
    """
    return loop().stop()


async def _switch():
    """
    A base async generator function used to create an awaitable coroutine
    for efficient async task switching.
    """
    while True:
        yield


# create an awaitable for efficient async task switching
switch = _switch().__aiter__().__anext__


__extras = {
    "__doc__": __doc__,
    "__main_exports__": __all__,
    "__package__": "aiootp",
    # "uvloop": uvloop,
    "asyncio": asyncio,
    "run": run,
    "time": time,
    "stop": stop,
    "loop": loop,
    "serve": serve,
    "sleep": sleep,
    "asleep": asleep,
    "switch": switch,
    "gather": gather,
    "new_task": new_task,
    "new_future": new_future,
    "default_loop": default_loop,
}


asynchs = commons.Namespace.make_module("asynchs", mapping=__extras)




import sys
import pytest
from pathlib import Path

PACKAGE_PATH = str(Path(__file__).absolute().parent.parent)
sys.path.append(PACKAGE_PATH)

from aiootp import *


TEST_STRING = "abcdefghijk"
TEST_STRING_LENGTH = len(TEST_STRING)
assert TEST_STRING_LENGTH == 11


@comprehension()
async def awith_return_coro(*, got=None):
    for _ in range(TEST_STRING_LENGTH):
        got = yield got
        await switch()
    raise UserWarning(got)


@comprehension()
def with_return_coro(*, got=None):
    for _ in range(TEST_STRING_LENGTH):
        got = yield got
    return got


@comprehension()
async def ano_return_coro(*, got=None):
    for _ in range(TEST_STRING_LENGTH):
        await switch()
        got = yield got


@comprehension()
def no_return_coro(*, got=None):
    for _ in range(TEST_STRING_LENGTH):
        got = yield got


@comprehension()
async def awith_return_iterator():
    for char in TEST_STRING:
        await switch()
        yield char
    raise UserWarning(char)


@comprehension()
def with_return_iterator():
    for char in TEST_STRING:
        yield char
    return char


@comprehension()
async def ano_return_iterator():
    for char in TEST_STRING:
        await switch()
        yield char


@comprehension()
def no_return_iterator(got=TEST_STRING):
    for char in TEST_STRING:
        yield char


def test_comprehension():
    @comprehension()
    def is_Comprende_generator():
        yield

    assert is_Comprende_generator().__class__ == Comprende

    @comprehension()
    async def is_async_Comprende_generator():
        yield

    assert is_async_Comprende_generator().__class__ == Comprende

    async def run_test_generators():
        for value in is_Comprende_generator():
            assert value == None

        async for value in is_async_Comprende_generator():
            assert value == None

    run(run_test_generators())


def test_comprehension_context():
    run(comprehension_context_testing())


async def comprehension_context_testing():
    async with awith_return_coro() as asend:
        assert await asend(None) == None
        for test in TEST_STRING:
            assert await asend(test) == test
    assert await asend.aresult() == "k", await asend.aresult()

    with with_return_coro() as send:
        assert send(None) == None
        for test in TEST_STRING:
            assert send(test) == test
    assert send.result() == "k", send.result()

    async with ano_return_coro() as asend:
        assert await asend(None) == None
        for test in TEST_STRING:
            assert await asend(test) == test
    try:
        failed = False
        return_value = await asend.aresult()
        assert return_value != None
    except AssertionError:
        failed = True
    finally:
        assert failed, return_value

    with no_return_coro() as send:
        assert send(None) == None
        for test in TEST_STRING:
            assert send(test) == test
    try:
        failed = False
        return_value = send.result()
        assert return_value != None
    except AssertionError:
        failed = True
    finally:
        assert failed, return_value


def test_comprehension_iter():
    run(comprehension_iteration_testing())


async def comprehension_iteration_testing():
    catcher = awith_return_iterator()
    async for index, value in catcher.atag():
        assert value == TEST_STRING[index], value
    return_value = await catcher.aresult()
    assert return_value == "k", return_value

    catcher = with_return_iterator()
    for index, value in catcher.tag():
        assert value == TEST_STRING[index], value
    return_value = catcher.result()
    assert return_value == "k", return_value

    catcher = ano_return_iterator()
    async for index, value in catcher.atag():
        assert value == TEST_STRING[index], value
    try:
        failed = False
        return_value = await catcher.aresult()
        assert return_value != None
    except AssertionError:
        failed = True
    finally:
        assert failed, return_value

    catcher = no_return_iterator()
    for index, value in catcher.tag():
        assert value == TEST_STRING[index], value
    try:
        failed = False
        return_value = catcher.result()
        assert return_value != None
    except AssertionError:
        failed = True
    finally:
        assert failed, return_value




# This file is part of aiootp, an asynchronous one-time-pad based crypto
# and anonymity library.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2020 Gonzo Investigatory Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2020 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


__all__ = [
    "generics",
    "Comprende",
    "comprehension",
    "azip",
    "anext",
    "arange",
    "aunpack",
    "unpack",
    "abirth",
    "birth",
    "adata",
    "data",
    "aorder",
    "order",
    "apick",
    "pick",
    "acycle",
    "cycle",
    "apopleft",
    "popleft",
    "ajson_decode",
    "json_decode",
    "ajson_encode",
    "json_encode",
    "asha_256",
    "sha_256",
    "asha_256_hmac",
    "sha_256_hmac",
    "asha_512",
    "sha_512",
    "asha_512_hmac",
    "sha_512_hmac",
    "anc_256",
    "nc_256",
    "anc_256_hmac",
    "nc_256_hmac",
    "anc_512",
    "nc_512",
    "anc_512_hmac",
    "nc_512_hmac",
]


__doc__ = """
A collection of basic utilities for simplifying & supporting the rest of
the codebase.
"""


import re
import json
import heapq
import random
import binascii
import pybase64
import builtins
import itertools
import aioitertools
from sys import getsizeof
from functools import wraps
from functools import lru_cache
from async_lru import alru_cache
from types import GeneratorType
from types import AsyncGeneratorType
from random import uniform
from sympy import isprime as is_prime
from contextlib import contextmanager
from hashlib import sha3_256
from hashlib import sha3_512
from collections import deque
from collections.abc import Iterable
from collections.abc import Iterator
from collections.abc import AsyncIterable
from collections.abc import AsyncIterator
from aiocontext import async_contextmanager
from inspect import getsource
from inspect import isawaitable as is_awaitable
from inspect import iscoroutinefunction as is_async_function
from inspect import isasyncgenfunction as is_async_gen_function
from inspect import isgeneratorfunction as is_generator_function
from .commons import *
from .asynchs import *
from .asynchs import time
from . import DEBUG_MODE


count = itertools.count
aiter = aioitertools.iter
acount = aioitertools.count


def src(obj, display=True):
    """
    Prints the source code of an object to the screen or, if ``display``
    is toggled to a falsey value, returns the source code instead.
    """
    if display:
        print(getsource(obj))
    else:
        return getsource(obj)


def size_of(obj, display=False):
    """
    Returns the memory size of an object `data` in bytes.
    """
    if not display:
        return getsizeof(obj)
    else:
        print(getsizeof(obj))


class AsyncInit(type):
    """
    A metaclass that allows classes to use asynchronous ``__init__``
    methods. Inspired by David Beazley.
    """
    async def __call__(cls, *args, **kwargs):
        self = cls.__new__(cls, *args, **kwargs)
        await self.__init__(*args, **kwargs)
        return self


class Enumerate:
    """
    An ``enumerate`` variant that supports sync & async generators.
    """
    __slots__ = ["gen", "start"]

    def __init__(self, gen, start=0):
        self.gen = gen
        self.start = start

    async def __aiter__(self):
        """
        Adds an incrementing number to each yielded result of either an
        async or sync generator.
        """
        if is_async_iterable(self.gen):
            counter = self.start
            async for result in self.gen:
                yield counter, result
                counter += 1
        else:
            for result in self.__iter__():
                await switch()
                yield result

    def __iter__(self):
        """
        Adds an incrementing number to each yielded result of either a
        synchronous generator.
        """
        counter = self.start
        for result in self.gen:
            yield counter, result
            counter += 1


class MemberFromStaticMethod:
    @staticmethod
    def convert(self, static_method_name, static_method, *args, **kwargs):
        @wraps(static_method)
        def wrapped_static_method(*a, **kw):
            try:
                new_args = list(args)
                for index, new_arg in enumerate(a):
                    new_args[index] = new_arg
                new_kwargs = dict(kwargs)
                new_kwargs.update(kw)
            except IndexError as error:
                raise NotImplementedError(
                    "Sending args into a method after they were replaced."
                ) from error
            return static_method(*new_args, **new_kwargs)

        setattr(self, static_method_name, wrapped_static_method)


def display_exception_info(error):
    """
    Prints out debug information of exceptions.
    """
    print("Error Type:", error)
    print("Error Args:", error.args)
    print("Error Cause:", error.__cause__)
    print("Error Value:", getattr(error, "value", None))


class ExampleException(Exception):
    """
    Empty, unused placeholder exception.
    """


class AsyncRelayExceptions:
    __slots__ = ["aexcept_code", "afinally_code"]

    read_me = f"""
    Overwrite {__slots__} methods with custom async functions.
    They will proc in ``aiootp.generics.aignore`` async context manager if:

    1.  {__slots__[0]} - when the ignored exceptions are raised
    within the context.

    But always,
    2.  {__slots__[1]} - at the end of the context.
    """

    def __init__(self):
        async def placeholder():
            return self.read_me

        self.aexcept_code = placeholder
        self.afinally_code = placeholder


class RelayExceptions:
    __slots__ = ["except_code", "finally_code"]

    read_me = f"""
    Overwrite {__slots__} methods with custom functions.
    They will proc in ``aiootp.generics.ignore`` context manager if:

    1.  {__slots__[0]} - when the ignored exceptions are raised
    within the context.

    But always,
    2.  {__slots__[1]} - at the end of the context.
    """

    def __init__(self):
        def placeholder():
            return self.read_me

        self.except_code = placeholder
        self.finally_code = placeholder


@async_contextmanager
async def aignore(*exceptions, display=False):
    """
    Usage example:

    async with aignore(TypeError):
        c = a + b
        # exception is surpressed if adding a and b raises a TypeError

    Or, dynamically choose which exceptions to catch, and call custom
    cleanup code. ->

    async def cleanup():
        await database.asave()

    async with ignore() as error_relay:
        error_relay.aexcept_code = cleanup
        # This will close ``database`` if either DynamicException or
        # IOError are raised within the block.

        error_relay.afinally_code = cleanup
        # This will ensure close is called on ``database`` in a finally
        # block.
    """
    try:
        exceptions = exceptions if exceptions else ExampleException
        relay = AsyncRelayExceptions()
        yield relay
    except exceptions as error:
        if display:
            display_exception_info(error)
        await relay.aexcept_code()
    except Exception as error:
        if display:
            display_exception_info(error)
        raise error
    finally:
        await relay.afinally_code()


@contextmanager
def ignore(*exceptions, display=False):
    """
    Usage example:

    with ignore(TypeError):
        c = a + b
        # exception is surpressed if adding a and b raises a TypeError

    Or, dynamically choose which exceptions to catch, and call custom
    cleanup code. ->

    def cleanup():
        database.save()

    with ignore(DynamicException, IOError) as error_relay:
        error_relay.except_code = cleanup
        # This will close ``database`` if either DynamicException or
        # IOError are raised within the block.

        error_relay.finally_code = cleanup
        # This will ensure close is called on ``database`` in a finally
        # block.
    """
    try:
        exceptions = exceptions if exceptions else ExampleException
        relay = RelayExceptions()
        yield relay
    except exceptions as error:
        relay.except_code()
        if display:
            display_exception_info(error)
    except Exception as error:
        if display:
            display_exception_info(error)
        raise error
    finally:
        relay.finally_code()


def is_exception(obj):
    """
    Returns a bool of whether ``obj`` is an exception object.
    """
    return hasattr(obj, "__cause__")


def is_async_iterable(obj):
    """
    Returns a bool of whether ``obj`` supports async iteration.
    """
    return isinstance(obj, AsyncIterable)


def is_iterable(obj):
    """
    Returns a bool of whether ``obj`` supports iteration.
    """
    return isinstance(obj, Iterable)


def is_async_iterator(obj):
    """
    Returns a bool of whether ``obj`` is an async iterator.
    """
    return isinstance(obj, AsyncIterator)


def is_iterator(obj):
    """
    Returns a bool of whether ``obj`` is an iterator.
    """
    return isinstance(obj, Iterator)


def is_async_generator(obj):
    """
    Returns a bool of whether ``obj`` is an async generator.
    """
    return isinstance(obj, AsyncGeneratorType)


def is_generator(obj):
    """
    Returns a bool of whether ``obj`` is an generator.
    """
    return isinstance(obj, GeneratorType)


def customize_parameters(a=(), kw=(), indexes=(), args=(), kwargs=()):
    """
    Replaces ``a`` and ``kw`` arguments & keyword arguments with ``args``
    if ``indexes`` is specified, and ``kwargs``.
    """
    if args and indexes:
        a = list(a)
        for index in indexes:
            a[index] = args[index]
    for kwarg in kwargs:
        kw[kwarg] = kwargs[kwarg]
    return a, kw


def comprehension(*args, indexes=(), catcher=None, **kwargs):
    """
    A decorator which wraps async & sync generator functions with the
    ``Comprende`` class, or the wrapper passed into ``catcher``. Can
    also optionally surgically replace arguments and keywords to the
    functions. This is helpful for being able to dynamically wrap
    functions in different contexts to alter behavior.
    """
    def func_catch(func):
        @wraps(func)
        def gen_wrapper(*a, **kw):
            nonlocal catcher

            a, kw = customize_parameters(a, kw, indexes, args, kwargs)
            catcher = Comprende if catcher == None else catcher
            return catcher(func, *a, **kw)

        return gen_wrapper

    return func_catch


class Comprende:
    """
    Comprende is a generator wrapper class that exposes an innovative,
    clean api for making sync & async generators more useful by making
    their many features easier to use. Comprende allows for easily
    retrieving generator "return" values, has built-in methods that
    support dotted chaining for inline data processing of generator
    outputs, & it opens channels of communication to and from sync &
    async coroutines designed to be driven by a caller.

    Comprende is what generator comprehensions should be.

    Usage Example:

    def gen(x=None, y=None):
        # The final result is returned normally in a sync generator ->
        z = yield x + y
        return x * y * z

    # Easily drive the generator forward.
    with Comprende(gen, x=1, y=2) as example:
        z = 3
        # This will send in ``None``
        sum_of_x_y = example()
        assert sum_of_x_y == 3

        # This will cause the generator to reach the return and exit ->
        example(z)

    # The result returned from the generator is now available ->
    product_of_x_y_z = example.result()
    assert product_of_x_y_z == 6

    # The ``example`` variable is actually the Comprende object, which
    # redirects values to the wrapped generator's ``send()`` method
    # using the instance's ``__call__()`` method.  It's still available
    # after the context closes ->
    assert example.__class__.__name__ == "Comprende"

    for result in Comprende(gen, 1, 2).sha_256():
        # This will hash each output of a generator and yield the hash.

    ciphertext = []
    key = aiootp.csprng()
    for result in Comprende(gen, 1, 2).str().encrypt(key):
        # This will stringify each output of a generator, then encrypt
        # each result, and yield the ciphertext.
        ciphertext.append(result)

    for result in aiootp.unpack(ciphertext).decrypt(key):
        # This will yield the original results in plaintext
        print(result)

    Async Usage Example:

    async def gen(x=None, y=None):
        # Because having a return statement in an async generator is a
        # SyntaxError, the return value is expected to be passed into
        # UserWarning, and then raised to propagate upstream. It's then
        # available from the instance's ``aresult`` method ->

        z = yield x + y
        result = x * y * z
        raise UserWarning(result)

    # Easily drive the generator forward.
    async with Comprende(gen, x=1, y=2) as example:
        z = 3
        # This will send in ``None``
        sum_of_x_y = await example()
        assert sum_of_x_y == 3

        # This will cause the generator to reach the return & exit the
        # context manager ->
        await example(z)

    # The result returned from the generator is now available ->
    product_of_x_y_z = await example.aresult()
    assert product_of_x_y_z == 6

    # The ``example`` variable is actually the Comprende object, which
    # redirects values to the wrapped generator's ``asend()`` method
    # using the instance's ``__call__()`` method. It's still available
    # after the context closes ->
    assert example.__class__.__name__ == "Comprende"

    async for result in Comprende(gen, 1, 2).asha_256():
        # This will hash each output of a generator and yield the hash.

    ciphertext = []
    key = await aiootp.acsprng()
    async for result in Comprende(gen, 1, 2).astr().aencrypt(key):
        # This will stringify each output of a generator, then encrypt
        # each result, and yield the ciphertext.
        ciphertext.append(result)

    async for result in aiootp.aunpack(ciphertext).adecrypt(key):
        # This will yield the original results in plaintext


    Comprende has many more useful features to play around with! Have
    fun with it!
    """
    decorator = comprehension

    _cached = {}

    __slots__ = [
        "gen",
        "func",
        "args",
        "kwargs",
        "iterator",
        "aencrypt",
        "encrypt",
        "adecrypt",
        "decrypt",
        "_async",
        "_runsum",
        "_return",
        "_thrown",
        "__call__",
        "_acache_yield",
        "_cache_yield",
    ]

    lazy_generators = {
        "_agetitem",
        "_getitem",
        "atimeout",
        "timeout",
        "ajson_loads",
        "json_loads",
        "ajson_dumps",
        "json_dumps",
        "aencode",
        "encode",
        "adecode",
        "decode",
        "astr",
        "str",
        "aint",
        "int",
        "ahex",
        "hex",
        "abin",
        "bin",
        "abytes",
        "bytes",
        "ato_base",
        "to_base",
        "afrom_base",
        "from_base",
        "azfill",
        "zfill",
        "aslice",
        "slice",
        "atag",
        "tag",
        "ahalt",
        "halt",
        "afeed",
        "feed",
        "afeed_self",
        "feed_self",
        "aresize",
        "resize",
        "ato_base64",
        "to_base64",
        "afrom_base64",
        "from_base64",
        "aint_to_ascii",
        "int_to_ascii",
        "aascii_to_int",
        "ascii_to_int",
        "asha_512",
        "sha_512",
        "asha_512_hmac",
        "sha_512_hmac",
        "asha_256",
        "sha_256",
        "asha_256_hmac",
        "sha_256_hmac",
    }

    eager_generators = {"aheappop", "heappop", "areversed", "reversed"}

    _generators = {"__aiter__", "__iter__"}

    lazy_methods = {"anext", "next", "asend", "send"}

    eager_methods = {
        "alist",
        "list",
        "adeque",
        "deque",
        "aset",
        "set",
        "adict",
        "dict",
        "ajoin",
        "join",
        "aexhaust",
        "exhaust",
    }

    _methods = {
        "athrows",
        "throws",
        "athrow",
        "throw",
        "aclose",
        "close",
        "areset",
        "reset",
        "aprime",
        "prime",
        "aresult",
        "result",
        "aclear",
        "clear",
        "acatch",
        "catch",
        "arelay",
        "relay",
        "acache_check",
        "cache_check",
        "aclass_clear",
        "class_clear",
        "runsum",
        "precomputed",
    }

    ASYNC_GEN_DONE = "async generator raised StopAsyncIteration"

    def __init__(self, func=None, *a, **kw):
        """
        Establishes async / sync properties of new objects & copies
        over wrapped functions' signatures.
        """
        self.args = a
        self.kwargs = kw
        self._runsum = ""
        self._thrown = deque()
        self._return = deque()
        if is_async_gen_function(func):
            self.func = func
            self.gen = self.__set_async()
            self.iterator = aiter(self.gen)
        else:
            self.func = func if func != None else unpack
            self.gen = self.__set_sync()
            self.iterator = iter(self.gen)

    async def areset(self):
        """
        Replaces the generator wrapper with a new async wrapper.
        """
        self.gen = self.__set_async()
        return self

    def reset(self):
        """
        Replaces the generator wrapper with a new sync wrapper.
        """
        self.gen = self.__set_sync()
        return self

    async def __aexamine_sent_exceptions(self, gen=None, got=None):
        """
        Catches ``UserWarning``s which signals that the generator, or a
        subgenerator in the stack, has raised a return value.
        """
        while True:
            got = yield got
            if issubclass(got.__class__, UserWarning):
                if any(got.args):
                    self._thrown.append(got.args[0])
                await gen.athrow(got)

    def __set_async(self):
        """
        Does the wrapping of user async generators to allow catching
        return values.
        """
        @wraps(self.func)
        async def _acomprehension(gen=None):
            catch_UserWarning = self.__aexamine_sent_exceptions(gen)
            await catch_UserWarning.asend(None)
            async with self.acatch():
                got = None
                while True:
                    got = yield await gen.asend(got)
                    await catch_UserWarning.asend(got)

        self._async = True
        self.__call__ = self._acall
        return _acomprehension(self.func(*self.args, **self.kwargs))

    def __examine_sent_exceptions(self, gen=None, got=None):
        """
        Catches ``UserWarning``s which signals that the generator, or a
        subgenerator in the stack, has raised a return value.
        """
        while True:
            got = yield got
            if issubclass(got.__class__, UserWarning):
                if any(got.args):
                    self._thrown.append(got.args[0])
                gen.throw(got)

    def __set_sync(self):
        """
        Does the wrapping of user generators to allow catching return
        values.
        """
        @wraps(self.func)
        def _comprehension(gen=None):
            catch_UserWarning = self.__examine_sent_exceptions(gen)
            catch_UserWarning.send(None)
            with self.catch():
                got = None
                while True:
                    got = yield gen.send(got)
                    catch_UserWarning.send(got)

        self._async = False
        self.__call__ = self._call
        return _comprehension(self.func(*self.args, **self.kwargs))

    @async_contextmanager
    async def acatch(self):
        """
        Handles catching the return values passed through exceptions
        from async generators & makes sure other errors are propagated
        correctly up to user code. Asynchronous generators don't already
        have a mechanism for returning values. So this async context
        manager handles catching return values from UserWarning
        exceptions & appends those results to ``self._return``. Items
        in the result queue are accessible from ``self.aresult()``.
        """
        try:
            yield self
        except GeneratorExit as done:
            raise done
        except RuntimeError as done:
            if self.ASYNC_GEN_DONE not in done.args:
                raise done
        except StopAsyncIteration:
            pass
        except UserWarning as done:
            if done.args:
                self._return.append(done.args[0])

    @contextmanager
    def catch(self):
        """
        Handles catching the return values passed through exceptions
        from sync generators & makes sure other errors are propagated
        correctly up to user code. Synchronous generators already have
        a mechanism for returning values. This context manager handles
        catching StopIteration values, & for the sake of parity with
        async generators, it also catches return values from UserWarning
        exceptions & appends those results to ``self._return``. Items
        in the result queue are accessible from ``self.result()``.
        """
        try:
            yield self
        except StopIteration as done:
            if getattr(done, "value", None) != None:
                self._return.append(done.value)
        except UserWarning as done:
            if done.args:
                self._return.append(done.args[0])

    @async_contextmanager
    async def arelay(self, result=None, source=None):
        """
        This is a lower level context manager for users who've created
        async generators that need to propagate results up to calling
        code. Code in this context manager's block will return ``result``
        or the return value of a ``source`` Comprende async generator
        up to its caller in a UserWarning exception.
        """
        try:
            source = source if source else self
            yield self
        except UserWarning:
            if result != None:
                raise UserWarning(result)
            raise UserWarning(await source.aresult(exit=True))

    @contextmanager
    def relay(self, result=None, source=None):
        """
        This is a lower level context manager for users who've created
        sync generators that need to propagate results up to calling
        code. Code in this context manager's block will relay a ``result``
        or the return value of a ``source`` Comprende sync generator
        up to its caller in a UserWarning exception.
        """
        try:
            source = source if source else self
            yield self
        except (StopIteration, UserWarning):
            if result != None:
                raise UserWarning(result)
            raise UserWarning(source.result(exit=True))

    async def aprime(self):
        """
        Resets the instance's async wrapper generator & ``asend``s in a
        ``None`` value to prime the generator, i.e. bringing it to the
        first yield statement.
        """
        await self.areset()
        await self(None)
        return self

    def prime(self):
        """
        Resets the instance's async wrapper generator & ``asend``s in a
        ``None`` value to prime the generator, i.e. bringing it to the
        first yield statement.
        """
        self.reset()
        self(None)
        return self

    async def asend(self, got=None):
        """
        Copies the wrapped async generator's ``asend`` method behavior.
        This is equivalent to an async Comprende generator's ``__call__``
        method.
        """
        return await self(got)

    def send(self, got=None):
        """
        Copies the wrapped sync generator's ``send`` method behavior.
        This is equivalent to a sync Comprende generator's ``__call__``
        method.
        """
        return self(got)

    async def athrow(self, *a, **kw):
        """
        This is quivalent to a wrapped async generator's ``athrow``
        method.
        """
        await self.gen.athrow(*a, **kw)
        return await self.aresult()

    def throw(self, *a, **kw):
        """
        This is quivalent to a wrapped sync generator's ``throw`` method.
        """
        self.gen.throw(*a, **kw)
        return self.result()

    async def aclose(self, *a, **kw):
        """
        This is quivalent to a wrapped async generator's ``aclose``
        method.
        """
        return await self.gen.aclose(*a, **kw)

    def close(self, *a, **kw):
        """
        This is quivalent to a wrapped sync generator's ``close`` method.
        """
        return self.gen.close(*a, **kw)

    async def aresult(self, *, pop=False, exit=False, silent=True):
        """
        Controls access to instance results. This method can cause an
        async generator to close when ``exit`` is truthy & returns its
        results if it has any. If ``pop`` is truthy, the results are
        popped off the result queue, & by default surpresses ``IndexError``
        if no results are in the queue. If ``silent`` is truthy, then
        exceptions are not ignored & the first raised exception is
        printed to stdout.
        """
        if exit and silent:
            with ignore(TypeError, display=not silent):
                await gather(self(UserWarning()), return_exceptions=True)
        elif exit:
            await self(UserWarning())
        with ignore(IndexError, display=not silent):
            if pop:
                return self._return.popleft()
            else:
                return self._return[0]

    def result(self, *, pop=False, exit=False, silent=True):
        """
        Controls access to instance results. This method can cause an
        sync generator to close when ``exit`` is truthy & returns its
        results if it has any. If ``pop`` is truthy, the results are
        popped off the result queue, & by default surpresses ``IndexError``
        if no results are in the queue. If ``silent`` is truthy, then
        exceptions are not ignored & the first raised exception is
        printed to stdout.
        """
        if exit and silent:
            with ignore(TypeError, StopIteration, display=not silent):
                self(UserWarning())
        elif exit:
            self(UserWarning())
        with ignore(IndexError, display=not silent):
            if pop:
                return self._return.popleft()
            else:
                return self._return[0]

    async def athrows(self):
        """
        When causing an async generator to exit using ``self.aresult()``,
        the exception is appended to the instance ``self._thrown`` queue.
        Normally this doesn't have significance, but users can adapt this
        queue & override ``self.aresult`` to send in relevant values
        instead.
        """
        return self._thrown

    def throws(self):
        """
        When causing a sync generator to exit using ``self.result()``,
        the exception is appended to the instance ``self._thrown`` queue.
        Normally this doesn't have significance, but users can adapt this
        queue & override ``self.result`` to send in relevant values
        instead.
        """
        return self._thrown

    async def anext(self):
        """
        Advances the wrapped async generator to the next yield.
        """
        return await self.iterator.asend(None)

    def next(self):
        """
        Advances the wrapped sync generator to the next yield.
        """
        return builtins.next(self.iterator)

    @classmethod
    async def aclear_class(cls):
        """
        Allows users to manually clear the cache of all the class'
        instances.
        """
        await cls().aclear(cls=True)

    @classmethod
    def clear_class(cls):
        """
        Allows users to manually clear the cache of all the class'
        instances.
        """
        cls().clear(cls=True)

    def __del__(self):
        """
        Attempts to cleanup instance caches when deleted or garbage
        collected to reduce memory overhead.
        """
        self.clear()
        if hasattr(self, "gen"):
            del self.gen

    async def aclear(self, *, cls=False):
        """
        Allows users to manually clear the cache of an instance, or if
        ``cls`` is ``True`` clears the cache of every instance.
        """
        try:
            if cls == True:
                for instance in dict(self.__class__._cached).values():
                    await instance.aclear()
            elif self.precomputed != False:
                del self.__class__._cached[self.runsum]
                async for cache in self._acache_has():
                    cache.cache_clear()
        finally:
            self._runsum = ""

    def clear(self, *, cls=False):
        """
        Allows users to manually clear the cache of an instance, or if
        ``cls`` is ``True`` clears the cache of every instance.
        """
        try:
            if cls == True:
                for instance in dict(self.__class__._cached).values():
                    instance.clear()
            elif self.precomputed != False:
                del self.__class__._cached[self.runsum]
                for cache in self._cache_has():
                    cache.cache_clear()
        finally:
            self._runsum = ""

    async def _acache_has(self):
        """
        Returns the lru cached methods of an instance in an iterable.
        """
        if hasattr(self, "_cache_yield"):
            await switch()
            yield self._cache_yield
        if hasattr(self, "_acache_yield"):
            await switch()
            yield self._acache_yield

    def _cache_has(self):
        """
        Returns the lru cached methods of an instance in an iterable.
        """
        if hasattr(self, "_cache_yield"):
            yield self._cache_yield
        if hasattr(self, "_acache_yield"):
            yield self._acache_yield

    async def _aset_cache(self):
        """
        Creates a per-instance function that returns precomputed results
        with lru cache turned on to save on speed for instances that have
        called async methods or generators which do eager computation of
        an async generator's entire result set.
        """
        @alru_cache(maxsize=2)
        async def _acache_yield(runsum=None):
            results = []
            async for result in self:
                results.append(result)
            return results

        self._runsum = await self._amake_runsum()
        self._acache_yield = _acache_yield

    def _set_cache(self):
        """
        Creates a per-instance function that returns precomputed results
        with lru cache turned on to save on speed for instances that have
        called sync methods or generators which do eager computation of
        a generator's entire result set.
        """
        @lru_cache(maxsize=2)
        def _cache_yield(runsum=None):
            results = []
            for result in self:
                results.append(result)
            return results

        self._runsum = self._make_runsum()
        self._cache_yield = _cache_yield

    @classmethod
    async def _amake_runsum(*args, low=bits[512], high=bits[513]):
        """
        Calculates a 512-bit pseudo-random hex string id to instances
        to mark themselves as having cached results.
        """
        return await asha_512(
            int(uniform(low, high)) % random.choice(primes[256]), *args
        )

    @classmethod
    def _make_runsum(*args, low=bits[512], high=bits[513]):
        """
        Calculates a 512-bit pseudo-random hex string id to instances
        to mark themselves as having cached results.
        """
        return sha_512(
            int(uniform(low, high)) % random.choice(primes[256]), *args
        )

    @property
    def runsum(self):
        """
        Returns an empty string if the instance generator has not cached
        any results. Returns the generator's 32-byte hex string id if it
        has.
        """
        return self._runsum[:32]

    @property
    def precomputed(self):
        """
        Checks the class' dictionary of cached flags for the generator's
        ``self.runsum`` id. Returns the instance if found, False if not.
        """
        if self._runsum and self.runsum in self.__class__._cached:
            return self.__class__._cached[self.runsum]
        else:
            return False

    @async_contextmanager
    async def acache_check(self):
        """
        Exhausts the underlying Comprende async generator & joins the
        results together in a ``list``, then ``alru_cache``'s the result
        & yields it as a context manager. Finally, adds the instance
        into the class' ``_cached`` dictionary to more easily find &
        manage the memory overhead of caching values.
        """
        try:
            if not self._runsum:
                await self._aset_cache()
            yield await self._acache_yield(self.runsum)
        finally:
            self.__class__._cached[self.runsum] = self

    @contextmanager
    def cache_check(self):
        """
        Exhausts the underlying Comprende async generator & joins the
        results together in a ``list``, then ``alru_cache``'s the result
        & yields it as a context manager. Finally, adds the instance
        into the class' ``_cached`` dictionary to more easily find &
        manage the memory overhead of caching values.
        """
        try:
            if not self._runsum:
                self._set_cache()
            yield self._cache_yield(self.runsum)
        finally:
            self.__class__._cached[self.runsum] = self

    async def alist(self):
        """
        Exhausts the underlying Comprende async generator & joins the
        results together in a ``list``, then ``alru_cache``'s the result
        & returns it.
        """
        async with self.acache_check() as results:
            return list(results)

    def list(self):
        """
        Exhausts the underlying Comprende sync generator & joins the
        results together in a ``list``, then ``alru_cache``'s the result
        & returns it.
        """
        with self.cache_check() as results:
            return list(results)

    async def adeque(self):
        """
        Exhausts the underlying Comprende async generator & joins the
        results together in a ``collections.deque``, then ``alru_cache``'s
        the result & returns it.
        """
        async with self.acache_check() as results:
            return deque(results)

    def deque(self):
        """
        Exhausts the underlying Comprende sync generator & joins the
        results together in a ``collections.deque``, then ``alru_cache``'s
        the result & returns it.
        """
        with self.cache_check() as results:
            return deque(results)

    async def aset(self):
        """
        Exhausts the underlying Comprende async generator & joins the
        results together in a ``set``, then ``alru_cache``'s the result
        & returns it.
        """
        async with self.acache_check() as results:
            return set(results)

    def set(self):
        """
        Exhausts the underlying Comprende sync generator & joins the
        results together in a ``set``, then ``alru_cache``'s the result
        & returns it.
        """
        with self.cache_check() as results:
            return set(results)

    async def adict(self):
        """
        Exhausts the underlying Comprende async generator & joins the
        results together in a ``dict``, then ``alru_cache``'s the result
        & returns it.
        """
        async with self.acache_check() as results:
            return dict(results)

    def dict(self):
        """
        Exhausts the underlying Comprende sync generator & joins the
        results together in a ``dict``, then ``lru_cache``'s the result
        & returns it.
        """
        with self.cache_check() as results:
            return dict(results)

    async def ajoin(self, on=""):
        """
        Exhausts the underlying Comprende async generator & joins the
        results together ``on`` the string that's passed, then
        ``alru_cache``'s the result & returns it.
        """
        async with self.acache_check() as results:
            return on.join(results)

    def join(self, on=""):
        """
        Exhausts the underlying Comprende sync generator & joins the
        results together ``on`` the string that's passed, then
        ``lru_cache``'s the result & returns it.
        """
        with self.cache_check() as results:
            return on.join(results)

    async def aexhaust(self):
        """
        Iterates over the entirety of the underlying Comprende async
        generator without yielding the results. Instead, it only returns
        the final yielded result.
        """
        async for result in self:
            pass
        if result:
            return result

    def exhaust(self):
        """
        Iterates over the entirety of the underlying Comprende sync
        generator without yielding the results. Instead, it only returns
        the final yielded result.
        """
        for result in self:
            pass
        if result:
            return result

    async def atimeout(self, delay=5):
        """
        Stops the instance's wrapped async generator after a ``delay``
        number of seconds. Can only cancel during times when ``self``
        async iteration has yielded control back to the caller.
        """
        current_time = time()
        async for result in self:
            if time() - current_time < delay:
                yield result
            else:
                break

    def timeout(self, delay=5):
        """
        Stops the instance's wrapped sync generator after a ``delay``
        number of seconds. Can only cancel during times when ``self``
        iteration has yielded control back to the caller.
        """
        current_time = time()
        for result in self:
            if time() - current_time < delay:
                yield result
            else:
                break

    async def ahalt(self, sentinel="", *, sentinels=()):
        """
        Takes a ``sentinel`` or iterable of ``sentinels`` & halts the
        underlying Comprende async generator if it yields any of those
        sentinels.
        """
        sentinels = set(sentinels) if sentinels else {sentinel}
        async for result in self:
            if result in sentinels:
                break
            yield result

    def halt(self, sentinel="", *, sentinels=()):
        """
        Takes a ``sentinel`` or iterable of ``sentinels`` & halts the
        underlying Comprende sync generator if it yields any of those
        sentinels.
        """
        sentinels = set(sentinels) if sentinels else {sentinel}
        for result in self:
            if result in sentinels:
                break
            yield result

    async def afeed(self, iterable=None):
        """
        Takes in an sync or async iterable & sends those values into an
        async coroutine which automates the process of driving an async
        generator which is expecting results from a caller.
        """
        yield await self(None)
        async for food in aunpack(iterable):
            yield await self(food)

    def feed(self, iterable=None):
        """
        Takes in an iterable & sends those values into a sync coroutine
        which automates the process of driving a generator which is
        expecting results from a caller.
        """
        yield self(None)
        for food in iterable:
            yield self(food)

    async def afeed_self(self):
        """
        Recursively feeds the results of an async generator back into
        itself as coroutine values for the ``asend`` function.
        """
        food = await self(None)
        yield food
        while True:
            food = await self(food)
            yield food

    def feed_self(self):
        """
        Recursively feeds the results of an generator back into itself
        as coroutine values for the ``send`` function.
        """
        food = self(None)
        yield food
        while True:
            food = self(food)
            yield food

    async def atag(self, of=None):
        """
        By default behaves like ``enumerate`` for each value yielded
        from the underlying Comprende async generator. Optionally,
        ``of`` can be passed a sync or async iterable & prepends those
        values to the generator's results.
        """
        names = of if of != None else acount()
        async for name, item in azip(names, self):
            yield name, item

    def tag(self, of=None):
        """
        By default behaves like ``enumerate`` for each value yielded
        from the underlying Comprende sync generator. Optionally, ``of``
        can be passed an iterable & prepends those values to the
        generator's results.
        """
        names = of if of != None else count()
        for name, item in zip(names, self):
            yield name, item

    async def aheappop(self, span=None, *, of=None):
        """
        Exhausts the underlying Comprende async generator upto ``span``
        number of iterations, then yields the results in sorted order
        based on the ``heapq.heappop`` function.
        """
        if of != None:
            async with aunpack(of)[:span] as accumulator:
                results = await accumulator.alist()
            heapq.heapify(results)
            async for result in self:
                try:
                    yield result, heapq.heappop(results)
                except IndexError:
                    break
        else:
            async with aunpack(self)[:span] as accumulator:
                results = await accumulator.alist()
            heapq.heapify(results)
            while True:
                try:
                    yield heapq.heappop(results)
                except IndexError:
                    break

    def heappop(self, span=None, *, of=None):
        """
        Exhausts the underlying Comprende sync generator upto ``span``
        number of iterations, then yields the results in sorted order
        based on the ``heapq.heappop`` function.
        """
        if of != None:
            with unpack(of)[:span] as accumulator:
                results = accumulator.list()
            heapq.heapify(results)
            for result in self:
                try:
                    yield result, heapq.heappop(results)
                except IndexError:
                    break
        else:
            with unpack(self)[:span] as accumulator:
                results = accumulator.list()
            heapq.heapify(results)
            while True:
                try:
                    yield heapq.heappop(results)
                except IndexError:
                    break

    async def areversed(self, span=None, *, of=None):
        """
        Exhausts the underlying Comprende async generator upto ``span``
        number of iterations, then yields the results in reversed order.
        """
        if of != None:
            async with unpack(of)[:span] as accumulator:
                results = await accumulator.adeque()
            async for prev, result in azip(self, reversed(results)):
                yield prev, result
        else:
            async with unpack(self)[:span] as accumulator:
                results = await accumulator.adeque()
            for result in reversed(results):
                await switch()
                yield result

    def reversed(self, span=None, *, of=None):
        """
        Exhausts the underlying Comprende sync generator upto ``span``
        number of iterations, then yields the results in reversed order.
        """
        if of != None:
            with unpack(of)[:span] as accumulator:
                results = accumulator.deque()
            for prev, result in zip(self, reversed(results)):
                yield prev, result
        else:
            with unpack(self)[:span] as accumulator:
                results = accumulator.deque()
            for result in reversed(results):
                yield result

    async def aresize(self, size=128, *, of=None):
        """
        Buffers the output from the underlying Comprende async generator
        to yield the results in chunks of length ``size``.
        """
        iterable_self = aiter(self)
        if of:
            new_source = aiter(of)
            result = await anext(new_source)
            while True:
                while len(result) >= size:
                    cache = result[size:]
                    yield await anext(iterable_self), result[:size]
                    result = cache
                try:
                    result += await anext(new_source)
                except StopAsyncIteration:
                    break
            if result:
                yield await anext(iterable_self), result
        else:
            result = await anext(iterable_self)
            while True:
                while len(result) >= size:
                    cache = result[size:]
                    yield result[:size]
                    result = cache
                try:
                    result += await anext(iterable_self)
                except StopAsyncIteration:
                    break
            if result:
                yield result

    def resize(self, size=128, *, of=None):
        """
        Buffers the output from the underlying Comprende sync generator
        to yield the results in chunks of length ``size``.
        """
        iterable_self = iter(self)
        if of:
            new_source = iter(of)
            result = next(new_source)
            while True:
                while len(result) >= size:
                    cache = result[size:]
                    yield next(iterable_self), result[:size]
                    result = cache
                try:
                    result += next(new_source)
                except StopIteration:
                    break
            if result:
                yield next(iterable_self), result
        else:
            result = next(iterable_self)
            while True:
                while len(result) >= size:
                    cache = result[size:]
                    yield result[:size]
                    result = cache
                try:
                    result += next(iterable_self)
                except StopIteration:
                    break
            if result:
                yield result

    async def ato_base64(self, *, of=None):
        """
        Applies ``pybase64.standard_b64encode`` conversion to each value
        that's yielded from the underlying Comprende async generator
        before yielding the result.
        """
        if of != None:
            async for prev, result in azip(self, of):
                yield prev, to_b64(result)
        else:
            async for result in self:
                yield to_b64(result)

    def to_base64(self, *, of=None):
        """
        Applies ``pybase64.standard_b64encode`` conversion to each value
        that's yielded from the underlying Comprende sync generator
        before yielding the result.
        """
        if of != None:
            for prev, result in zip(self, of):
                yield prev, to_b64(result)
        else:
            for result in self:
                yield to_b64(result)

    async def afrom_base64(self, *, of=None):
        """
        Applies ``pybase64.standard_b64decode`` conversion to each value
        that's yielded from the underlying Comprende async generator
        before yielding the result.
        """
        if of != None:
            async for prev, result in azip(self, of):
                yield prev, from_b64(result)
        else:
            async for result in self:
                yield from_b64(result)

    def from_base64(self, *, of=None):
        """
        Applies ``pybase64.standard_b64decode`` conversion to each value
        that's yielded from the underlying Comprende sync generator
        before yielding the result.
        """
        if of != None:
            for prev, result in zip(self, of):
                yield prev, from_b64(result)
        else:
            for result in self:
                yield from_b64(result)

    async def aint_to_ascii(self, *, of=None):
        """
        Applies a ``binascii`` int-to-ascii conversion to each value
        that's yielded from the underlying Comprende async generator
        before yielding the result.
        """
        if of != None:
            async for prev, result in azip(self, of):
                yield prev, result.to_bytes(
                    (result.bit_length() + 7) // 8, "big"
                ).decode()
        else:
            async for result in self:
                yield result.to_bytes(
                    (result.bit_length() + 7) // 8, "big"
                ).decode()

    def int_to_ascii(self, *, of=None):
        """
        Applies a ``binascii`` int-to-ascii conversion to each value
        that's yielded from the underlying Comprende sync generator
        before yielding the result.
        """
        if of != None:
            for prev, result in azip(self, of):
                yield prev, result.to_bytes(
                    (result.bit_length() + 7) // 8, "big"
                ).decode()
        else:
            for result in self:
                yield result.to_bytes(
                    (result.bit_length() + 7) // 8, "big"
                ).decode()

    async def aascii_to_int(self, *, of=None):
        """
        Applies a ``binascii`` ascii-to-int conversion to each value
        that's yielded from the underlying Comprende async generator
        before yielding the result.
        """
        if of != None:
            async for prev, result in azip(self, of):
                yield prev, int.from_bytes(result.encode(), "big")
        else:
            async for result in self:
                yield int.from_bytes(result.encode(), "big")

    def ascii_to_int(self, *, of=None):
        """
        Applies a ``binascii`` ascii-to-int conversion to each value
        that's yielded from the underlying Comprende sync generator
        before yielding the result.
        """
        if of != None:
            for prev, result in zip(self, of):
                yield prev, int.from_bytes(result.encode(), "big")
        else:
            for result in self:
                yield int.from_bytes(result.encode(), "big")

    async def asha_512(self, salt=None, *, of=None):
        """
        Applies ``hashlib.sha3_512()`` to each value that's yielded
        from the underlying Comprende async generator before yielding
        the result.
        """
        if salt:
            if of != None:
                async for prev, result in azip(self, of):
                    yield prev, await asha_512(result, salt)
            else:
                async for result in self:
                    yield await asha_512(result, salt)
        else:
            if of != None:
                async for prev, result in azip(self, of):
                    yield prev, await asha_512(result)
            else:
                async for result in self:
                    yield await asha_512(result)

    def sha_512(self, salt=None, *, of=None):
        """
        Applies ``hashlib.sha3_512()`` to each value that's yielded
        from the underlying Comprende sync generator before yielding
        the result.
        """
        if salt:
            if of != None:
                for prev, result in zip(self, of):
                    yield prev, sha_512(result, salt)
            else:
                for result in self:
                    yield sha_512(result, salt)
        else:
            if of != None:
                for prev, result in zip(self, of):
                    yield prev, sha_512(result)
            else:
                for result in self:
                    yield sha_512(result)

    async def asha_512_hmac(self, key=None, salt=None, *, of=None):
        """
        Applies a ``hashlib.sha3_512()`` based hmac algorithm to each
        value that's yielded from the underlying Comprende async generator
        before yielding the result.
        """
        if salt:
            if of != None:
                async for prev, result in azip(self, of):
                    yield prev, await asha_512_hmac((result, salt), key=key)
            else:
                async for result in self:
                    yield await asha_512_hmac((result, salt), key=key)
        else:
            if of != None:
                async for prev, result in azip(self, of):
                    yield prev, await asha_512_hmac(result, key=key)
            else:
                async for result in self:
                    yield await asha_512_hmac(result, key=key)

    def sha_512_hmac(self, key=None, salt=None, *, of=None):
        """
        Applies a ``hashlib.sha3_512()`` based hmac algorithm to each
        value that's yielded from the underlying Comprende sync generator
        before yielding the result.
        """
        if salt:
            if of != None:
                for prev, result in zip(self, of):
                    yield prev, sha_512_hmac((result, salt), key=key)
            else:
                for result in self:
                    yield sha_512_hmac((result, salt), key=key)
        else:
            if of != None:
                for prev, result in zip(self, of):
                    yield prev, sha_512_hmac(result, key=key)
            else:
                for result in self:
                    yield sha_512_hmac(result, key=key)

    async def asha_256(self, salt=None, *, of=None):
        """
        Applies ``hashlib.sha3_256()`` to each value that's yielded
        from the underlying Comprende async generator before yielding
        the result.
        """
        if salt:
            if of != None:
                async for prev, result in azip(self, of):
                    yield prev, await asha_256(result, salt)
            else:
                async for result in self:
                    yield await asha_256(result, salt)
        else:
            if of != None:
                async for prev, result in azip(self, of):
                    yield prev, await asha_256(result)
            else:
                async for result in self:
                    yield await asha_256(result)

    def sha_256(self, salt=None, *, of=None):
        """
        Applies ``hashlib.sha3_256()`` to each value that's yielded
        from the underlying Comprende sync generator before yielding the
        result.
        """
        if salt:
            if of != None:
                for prev, result in zip(self, of):
                    yield prev, sha_256(result, salt)
            else:
                for result in self:
                    yield sha_256(result, salt)
        else:
            if of != None:
                for prev, result in zip(self, of):
                    yield prev, sha_256(result)
            else:
                for result in self:
                    yield sha_256(result)

    async def asha_256_hmac(self, key=None, salt=None, *, of=None):
        """
        Applies a ``hashlib.sha3_256()`` based hmac algorithm to each
        value that's yielded from the underlying Comprende async generator
        before yielding the result.
        """
        if salt:
            if of != None:
                async for prev, result in azip(self, of):
                    yield prev, await asha_256_hmac((result, salt), key=key)
            else:
                async for result in self:
                    yield await asha_256_hmac((result, salt), key=key)
        else:
            if of != None:
                async for prev, result in azip(self, of):
                    yield prev, await asha_256_hmac(result, key=key)
            else:
                async for result in self:
                    yield await asha_256_hmac(result, key=key)

    def sha_256_hmac(self, key=None, salt=None, *, of=None):
        """
        Applies a ``hashlib.sha3_256()`` based hmac algorithm to each
        value that's yielded from the underlying Comprende sync generator
        before yielding the result.
        """
        if salt:
            if of != None:
                for prev, result in zip(self, of):
                    yield prev, sha_256_hmac((result, salt), key=key)
            else:
                for result in self:
                    yield sha_256_hmac((result, salt), key=key)
        else:
            if of != None:
                for prev, result in zip(self, of):
                    yield prev, sha_256_hmac(result, key=key)
            else:
                for result in self:
                    yield sha_256_hmac(result, key=key)

    async def aint(self, *a, of=None, **kw):
        """
        Applies ``builtins.int(*a)`` to each value that's yielded from
        the underlying Comprende async generator before yielding the
        result.
        """
        if of != None:
            async for prev, result in azip(self, of):
                yield prev, int(result, *a, **kw)
        else:
            async for result in self:
                yield int(result, *a, **kw)

    def int(self, *a, of=None, **kw):
        """
        Applies ``builtins.int(*a)`` to each value that's yielded from
        the underlying Comprende sync generator before yielding the
        result.
        """
        if of != None:
            for prev, result in zip(self, of):
                yield prev, builtins.int(result, *a, **kw)
        else:
            for result in self:
                yield builtins.int(result, *a, **kw)

    async def ato_base(self, base=16, table=ASCII_ALPHANUMERIC, *, of=None):
        """
        Converts each integer value that's yielded from the underlying
        Comprende async generator to a string in ``base`` before yielding
        the result.
        """
        if of != None:
            async for prev, result in azip(self, of):
                yield prev, await ainverse_int(result, base, table)
        else:
            async for result in self:
                yield await ainverse_int(result, base, table)

    def to_base(self, base=16, table=ASCII_ALPHANUMERIC, *, of=None):
        """
        Converts each integer value that's yielded from the underlying
        Comprende sync generator to a string in ``base`` before yielding
        the result.
        """
        if of != None:
            for prev, result in zip(self, of):
                yield prev, inverse_int(result, base, table)
        else:
            for result in self:
                yield inverse_int(result, base, table)

    async def afrom_base(self, base, table=ASCII_ALPHANUMERIC, *, of=None):
        """
        Convert string results of generator results in numerical ``base``
        into decimal.
        """
        if of != None:
            async for prev, result in azip(self, of):
                yield prev, await abase_to_decimal(result, base, table)
        else:
            async for result in self:
                yield await abase_to_decimal(result, base, table)

    def from_base(self, base, table=ASCII_ALPHANUMERIC, *, of=None):
        """
        Convert ``string`` in numerical ``base`` into decimal.
        """
        if of != None:
            for prev, result in zip(self, of):
                yield prev, base_to_decimal(result, base, table)
        else:
            for result in self:
                yield base_to_decimal(result, base, table)

    async def azfill(self, *a, of=None, **kw):
        """
        Applies ``builtins.zfill(*a)`` to each value that's yielded from
        the underlying Comprende async generator before yielding the
        result.
        """
        if of != None:
            async for prev, result in azip(self, of):
                yield prev, result.zfill(*a, **kw)
        else:
            async for result in self:
                yield result.zfill(*a, **kw)

    def zfill(self, *a, of=None, **kw):
        """
        Applies ``builtins.zfill(*a)`` to each value that's yielded from
        the underlying Comprende sync generator before yielding the
        result.
        """
        if of != None:
            for prev, result in zip(self, of):
                yield prev, result.zfill(*a, **kw)
        else:
            for result in self:
                yield result.zfill(*a, **kw)

    async def aslice(self, *a, of=None):
        """
        Applies ``builtins.slice(*a)`` to each value that's yielded from
        the underlying Comprende async generator before yielding the
        result.
        """
        if of != None:
            async for prev, result in azip(self, of):
                yield prev, result[slice(*a)]
        else:
            async for result in self:
                yield result[slice(*a)]

    def slice(self, *a, of=None):
        """
        Applies ``builtins.slice(*a)`` to each value that's yielded from
        the underlying Comprende sync generator before yielding the
        result.
        """
        if of != None:
            for prev, result in zip(self, of):
                yield prev, result[slice(*a)]
        else:
            for result in self:
                yield result[slice(*a)]

    async def astr(self, *a, of=None, **kw):
        """
        Applies ``builtins.str(*a)`` to each value that's yielded from
        the underlying Comprende async generator before yielding the
        result.
        """
        if of != None:
            async for prev, result in azip(self, of):
                yield prev, str(result, *a, **kw)
        else:
            async for result in self:
                yield str(result, *a, **kw)

    def str(self, *a, of=None, **kw):
        """
        Applies ``builtins.str()`` to each value that's yielded from
        the underlying Comprende sync generator before yielding the
        result.
        """
        if of != None:
            for prev, result in zip(self, of):
                yield prev, builtins.str(result, *a, **kw)
        else:
            for result in self:
                yield builtins.str(result, *a, **kw)

    async def aencode(self, *a, of=None, **kw):
        """
        Applies ``builtins.str(*a)`` to each value that's yielded from
        the underlying Comprende async generator before yielding the
        result.
        """
        if of != None:
            async for prev, result in azip(self, of):
                yield prev, result.encode(*a, **kw)
        else:
            async for result in self:
                yield result.encode(*a, **kw)

    def encode(self, *a, of=None, **kw):
        """
        Applies ``builtins.str()`` to each value that's yielded from
        the underlying Comprende sync generator before yielding the
        result.
        """
        if of != None:
            for prev, result in zip(self, of):
                yield prev, result.encode(*a, **kw)
        else:
            for result in self:
                yield result.encode(*a, **kw)

    async def adecode(self, *a, of=None, **kw):
        """
        Applies ``builtins.str(*a)`` to each value that's yielded from
        the underlying Comprende async generator before yielding the
        result.
        """
        if of != None:
            async for prev, result in azip(self, of):
                yield prev, result.decode(*a, **kw)
        else:
            async for result in self:
                yield result.decode(*a, **kw)

    def decode(self, *a, of=None, **kw):
        """
        Applies ``builtins.str()`` to each value that's yielded from
        the underlying Comprende sync generator before yielding the
        result.
        """
        if of != None:
            for prev, result in zip(self, of):
                yield prev, result.decode(*a, **kw)
        else:
            for result in self:
                yield result.decode(*a, **kw)

    async def ajson_loads(self, *a, of=None, **kw):
        """
        Applies ``json.loads()`` to each value that's yielded from
        the underlying Comprende async generator before yielding the
        result.
        """
        if of != None:
            async for prev, result in azip(self, of):
                yield prev, json.loads(result, *a, **kw)
        else:
            async for result in self:
                yield json.loads(result, *a, **kw)

    def json_loads(self, *a, of=None, **kw):
        """
        Applies ``json.loads()`` to each value that's yielded from
        the underlying Comprende sync generator before yielding the
        result.
        """
        if of != None:
            for prev, result in zip(self, of):
                yield prev, json.loads(result, *a, **kw)
        else:
            for result in self:
                yield json.loads(result, *a, **kw)

    async def ajson_dumps(self, *a, of=None, **kw):
        """
        Applies ``json.dumps()`` to each value that's yielded from
        the underlying Comprende async generator before yielding the
        result.
        """
        if of != None:
            async for prev, result in azip(self, of):
                yield prev, json.dumps(result, *a, **kw)
        else:
            async for result in self:
                yield json.dumps(result, *a, **kw)

    def json_dumps(self, *a, of=None, **kw):
        """
        Applies ``json.dumps()`` to each value that's yielded from
        the underlying Comprende sync generator before yielding the
        result.
        """
        if of != None:
            for prev, result in zip(self, of):
                yield prev, json.dumps(result, *a, **kw)
        else:
            for result in self:
                yield json.dumps(result, *a, **kw)

    async def ahex(self, *a, of=None, **kw):
        """
        Applies ``builtins.hex()`` to each value that's yielded from
        the underlying Comprende async generator before yielding the
        result.
        """
        if of != None:
            async for prev, result in azip(self, of):
                yield prev, hex(result, *a, **kw)
        else:
            async for result in self:
                yield hex(result, *a, **kw)

    def hex(self, *a, of=None, **kw):
        """
        Applies ``builtins.hex()`` to each value that's yielded from
        the underlying Comprende sync generator before yielding the
        result.
        """
        if of != None:
            for prev, result in zip(self, of):
                yield prev, builtins.hex(result, *a, **kw)
        else:
            for result in self:
                yield builtins.hex(result, *a, **kw)

    async def abytes(self, *a, of=None, **kw):
        """
        Applies ``builtins.bytes()`` to each value that's yielded from
        the underlying Comprende async generator before yielding the
        result.
        """
        if of != None:
            async for prev, result in azip(self, of):
                yield prev, bytes(result, *a, **kw)
        else:
            async for result in self:
                yield bytes(result, *a, **kw)

    def bytes(self, *a, of=None, **kw):
        """
        Applies ``builtins.bytes()`` to each value that's yielded from
        the underlying Comprende sync generator before yielding the
        result.
        """
        if of != None:
            for prev, result in zip(self, of):
                yield prev, builtins.bytes(result, *a, **kw)
        else:
            for result in self:
                yield builtins.bytes(result, *a, **kw)

    async def abin(self, *a, of=None, **kw):
        """
        Applies ``builtins.bin()`` to each value that's yielded from
        the underlying Comprende async generator before yielding the
        result.
        """
        if of != None:
            async for prev, result in azip(self, of):
                yield prev, bin(result, *a, **kw)
        else:
            async for result in self:
                yield bin(result, *a, **kw)

    def bin(self, *a, of=None, **kw):
        """
        Applies ``builtins.bin()`` to each value that's yielded from
        the underlying Comprende sync generator before yielding the
        result.
        """
        if of != None:
            for prev, result in zip(self, of):
                yield prev, builtins.bin(result, *a, **kw)
        else:
            for result in self:
                yield builtins.bin(result, *a, **kw)

    async def __aenter__(self):
        """
        Opens a context & gives the ``self`` object & its ``aresult``
        method in a tuple. Async generators / coroutines can be designed
        to return values by raising UserWarning(return_value).
        """
        await self.areset()
        return self

    def __enter__(self):
        """
        Opens a context & gives the ``self`` object & its ``result``
        method in a tuple. Results can be designed to return values
        by using ``return`` within the generator, or raising
        UserWarning(return_value).
        """
        self.reset()
        return self

    async def __aexit__(
        self, exc_type=None, exc_value=None, traceback=None
    ):
        """
        Surpresses StopAsyncIteration exceptions within a context.
        Clears the cached results upon exit.
        """
        try:
            if exc_type == StopAsyncIteration:
                return True
        finally:
            await self.aclear()

    def __exit__(self, exc_type=None, exc_value=None, traceback=None):
        """
        Surpresses StopIteration exceptions within a context. Clears the
        cached results upon exit.
        """
        try:
            if exc_type == StopIteration:
                return True
        finally:
            self.clear()

    async def __aiter__(self, *, got=None):
        """
        Iterates over the wrapped async &/or sync generator / coroutine
        & produces its values directly, or from lru_cache if an eager
        calculation has already computed the gererators values.
        """
        await self.areset()
        if self.precomputed != False:
            async with self.acache_check() as results:
                for result in results:
                    await switch()
                    yield result
        else:
            while True:
                try:
                    got = yield await self(got)
                except StopAsyncIteration:
                    break

    def __iter__(self, *, got=None):
        """
        Iterates over the wrapped generator / coroutine and produces its
        values directly, or from lru_cache if an eager calculation has
        already computed the gererators values.
        """
        self.reset()
        if self.precomputed != False:
            with self.cache_check() as results:
                for result in results:
                    yield result
        else:
            while True:
                try:
                    got = yield self(got)
                except StopIteration:
                    break

    def __next__(self):
        """
        Allows calling ``builtins.next`` on async / sync generators &
        coroutines.
        """
        if self._async:
            return self.anext()
        else:
            return self.next()

    async def _acall(self, got=None):
        """
        Allows the wrapped async generator / coroutine to receive
        ``asend`` values by using the class' __call__ method.
        """
        return await self.gen.asend(got)

    def _call(self, got=None):
        """
        Allows the wrapped generator / coroutine to receive ``send``
        values by using the class' __call__ method.
        """
        return self.gen.send(got)

    def __reversed__(self):
        """
        Allows reversing async/sync generators, but must compute all
        values first to do so.
        """
        if self._async:
            return self.areversed()
        else:
            return self.reversed()

    def __repr__(self, debugging=DEBUG_MODE):
        """
        Displays the string which, if ``exec``'d, would yield a new
        equivalent object.
        """
        a = self.args
        kw = self.kwargs
        func = self.func.__qualname__
        cls = self.__class__.__qualname__
        tab = f"\n{4 * ' '}"
        _repr = f"{cls}({tab}func={func},{tab}*{a},{tab}**{kw},\n)"
        if not debugging:
            key_finder = re.compile(r"[0-9a-fA-F]{64,}")
            for key in key_finder.finditer(_repr):
                _repr = _repr.replace(key.group(0), "<omitted-key>")
        return _repr

    def _set_index(self, index, spanner=builtins.range, _max=bits[256]):
        """
        Interprets the slice or int passed into __getitem__ into a
        range object.
        """
        if isinstance(index, int):
            step = 1
            start = index
            stop = index + 1
        else:
            step = index.step if isinstance(index.step, int) else 1
            start = index.start if isinstance(index.start, int) else 0
            stop = index.stop if isinstance(index.stop, int) else _max
        return start, stop, step, spanner(start, stop, step)

    async def _agetitem(self, index):
        """
        Allows indexing of async generators to yield the values
        associated with the slice or integer passed into the brackets.
        """
        start, stop, step, span = self._set_index(index, arange)
        with ignore(StopAsyncIteration):
            target = await span.anext()
            async for match, result in self.atag():
                if match >= stop:
                    break
                elif target == match:
                    yield result
                    target = await span.anext()

    def _getitem(self, index):
        """
        Allows indexing of generators to yield the values associated
        with the slice or integer passed into the brackets.
        """
        start, stop, step, span = self._set_index(index)
        limit = iter(span)
        with ignore(StopIteration):
            target = next(limit)
            for match, result in enumerate(self):
                if match >= stop:
                    break
                if target == match:
                    yield result
                    target = next(limit)

    def __getitem__(self, index):
        """
        Allows indexing of generators & async generators to yield the
        values associated with the slice or integer passed into the
        brackets.
        """
        if self._async:
            return self._agetitem(index)
        else:
            return self._getitem(index)

    for method in lazy_generators.union(eager_generators):
        vars()[method] = comprehension()(vars()[method])


def anext(coro_iterator):
    """
    Creates an asynchronous version of the ``builtins.next`` function.
    """
    try:
        return coro_iterator.__anext__()
    except AttributeError:
        return aiter(coro_iterator).__anext__()


@comprehension()
async def azip(*coros):
    """
    Creates an asynchronous version of the ``builtins.zip`` function.
    """
    tasks = deque()
    coros = [aiter(coro).__anext__ for coro in coros]
    try:
        while True:
            for coro in coros:
                await aappend(tasks, new_task(coro()))
            yield await gather(*tasks)
            tasks.clear()
    except StopAsyncIteration:
        pass


@comprehension()
def _zip(*iterables):
    """
    Creates a synchronous version of the zip builtin function which is
    wrapped by the ``Comprende`` class.
    """
    for results in zip(*iterables):
        yield results


@comprehension()
async def _acount(*args, **kwargs):
    """
    Creates an asynchronous version of ``aioitertools.count`` which is
    wrapped by the ``Comprende`` class.
    """
    async for result in acount(*args, **kwargs):
        yield result


@comprehension()
def _count(*args, **kwargs):
    """
    Creates a synchronous version of ``itertools.count`` that is wrapped
    by the ``Comprende`` class.
    """
    for result in count(*args, **kwargs):
        yield result


@comprehension()
async def _aiter(*a, **kw):
    """
    Creates an asynchronous version of ``aioitertools.iter`` which is
    wrapped by the ``Comprende`` class.
    """
    async for result in aiter(*a, **kw):
        yield result


@comprehension()
def _iter(*a, **kw):
    """
    Creates an asynchronous version of ``builtins.iter`` that is wrapped
    by the ``Comprende`` class.
    """
    for result in iter(*a, **kw):
        yield result


@comprehension()
async def acycle(iterable):
    """
    Unendingly cycles in order over the elements of an async iterable.
    """
    results = []
    async for result in iterable:
        yield result
        results.append(result)
    while True:
        for result in results:
            yield result


@comprehension()
def cycle(iterable):
    """
    Unendingly cycles in order over the elements of a sync iterable.
    """
    results = []
    for result in iterable:
        yield result
        results.append(result)
    while True:
        for result in results:
            yield result


@comprehension()
async def aunpack(iterable=None):
    """
    Runs through an iterable &/or async iterable & yields elements one
    at a time.
    """
    if iterable == None:
        iterable = acount()
    if is_async_iterable(iterable):
        async for item in iterable:
            yield item
    else:
        for item in iterable:
            await switch()
            yield item


@comprehension()
def unpack(iterable=None):
    """
    Runs through an iterable & yields elements one at a time.
    """
    if iterable == None:
        iterable = count()
    for result in iterable:
        yield result


@comprehension()
async def abirth(base="", *, stop=True):
    """
    Yields ``base`` ``times`` number of times. Useful for feeding other
    ``Comprende`` generators the totality of the value ``base``, which
    doesn't have to be either iterable or async iterable.
    """
    if stop:
        yield base
    else:
        while True:
            yield base


@comprehension()
def birth(base="", *, stop=True):
    """
    Yields ``base`` ``times`` number of times. Useful for feeding other
    ``Comprende`` generators the totality of the value ``base``, which
    doesn't have to be iterable.
    """
    if stop:
        yield base
    else:
        while True:
            yield base


@comprehension()
async def adata(sequence="", size=246, *, stop="__length_end__"):
    """
    Runs through a sequence & yields ``size`` sized chunks of the
    sequence one chunk at a time.
    """
    if stop == "__length_end__":
        stop = len(sequence) + size
    async for last, end in azip(
        arange(0, stop, size), arange(size, stop, size)
    ):
        yield sequence[last:end]


@comprehension()
def data(sequence="", size=246, *, stop="__length_end__"):
    """
    Runs through a sequence & yields ``size`` sized chunks of the
    sequence one chunk at a time.
    """
    if stop == "__length_end__":
        stop = len(sequence) + size
    for last, end in zip(range(0, stop, size), range(size, stop, size)):
        yield sequence[last:end]


@comprehension()
async def ajson_encode(raw_data=None, size=246):
    """
    Turns the ``json.dumps`` function into an async generator yielding
    ``size`` length chunks of string data per iteration.
    """
    async for result in adata(json.dumps(raw_data), size=size):
        yield result


@comprehension()
def json_encode(raw_data=None, size=246):
    """
    Turns the ``json.dumps`` function into a generator yielding ``size``
    length chunks of string data per iteration.
    """
    for result in data(json.dumps(raw_data), size=size):
        yield result


@comprehension()
async def ajson_decode(json_data=None):
    """
    Turns the ``json.loads`` function into an async generator yielding
    the data back in one iteration.
    """
    yield json.loads(json_data)


@comprehension()
def json_decode(json_data=None):
    """
    Turns the ``json.loads`` function into a generator yielding the data
    back in one iteration.
    """
    yield json.loads(json_data)


@comprehension()
async def aorder(*iterables):
    """
    Takes a collection of iterables &/or async iterables & exhausts them
    one at a time from left to right.
    """
    for iterable in iterables:
        async for result in aunpack(iterable):
            yield result


@comprehension()
def order(*iterables):
    """
    Takes a collection of iterables & exhausts them one at a time from
    left to right.
    """
    for iterable in iterables:
        for result in iterable:
            yield result


@comprehension()
async def askip(iterable, steps=1):
    """
    An async generator that produces the values yielded from ``iterable``
    once every ``steps`` number of iterations, otherwise produces
    ``None`` until ``iterable`` is exhausted.
    """
    async for result in iterable:
        for _ in range(steps):
            yield
        yield result


@comprehension()
def skip(iterable, steps=1):
    """
    A sync generator that produces the values yielded from ``iterable``
    once every ``steps`` number of iterations, otherwise produces
    ``None`` until ``iterable`` is exhausted.
    """
    for result in iterable:
        for _ in range(steps):
            yield
        yield result


@comprehension()
async def acompact(iterable, batch_size=1):
    """
    An async generator that yields ``batch_size`` number of elements
    from an async or sync ``iterable`` at a time.
    """
    stack = commons.Namespace()
    indexes = list(reversed(range(batch_size)))
    async for toggle, item in azip(cycle(indexes), iterable):
        stack[toggle] = item
        if not toggle:
            yield list(stack.namespace.values())
            stack.namespace.clear()
    if len(stack.namespace) < batch_size:
        yield list(stack.namespace.values())


@comprehension()
def compact(iterable, batch_size=1):
    """
    A generator that yields ``batch_size`` number of elements from an
    ``iterable`` at a time.
    """
    stack = commons.Namespace()
    indexes = list(reversed(range(batch_size)))
    for toggle, item in zip(cycle(indexes), iterable):
        stack[toggle] = item
        if not toggle:
            yield list(stack.namespace.values())
            stack.namespace.clear()
    if len(stack.namespace) < batch_size:
        yield list(stack.namespace.values())


@comprehension()
async def apopleft(queue):
    """
    An async generator which calls the ``popleft()`` method on ``queue``
    for every iterations, & exits on ``IndexError``.
    """
    while True:
        try:
            yield queue.popleft()
        except IndexError:
            break


@comprehension()
def popleft(queue):
    """
    A generator which calls the ``popleft()`` method on ``queue`` for
    every iterations, & exits on ``IndexError``.
    """
    while True:
        try:
            yield queue.popleft()
        except IndexError:
            break


@comprehension()
async def apick(names=None, mapping=None):
    """
    Does a bracketed lookup on ``mapping`` for each name in ``names``.
    """
    async for name in names:
        try:
            yield mapping[name]
        except KeyError:
            break


@comprehension()
def pick(names=None, mapping=None):
    """
    Does a bracketed lookup on ``mapping`` for each name in ``names``.
    """
    for name in names:
        try:
            yield mapping[name]
        except KeyError:
            break


@comprehension()
async def atimeout(iterable, delay=None):
    """
    Iterates over an async or sync ``iterable`` & halts iteration after
    ``delay`` time, even if results haven't been exhausted. Can only
    cancel during times when ``iterable`` has yielded control back to
    the caller.
    """
    async for result in aunpack(iterable).atimeout(delay):
        yield result


@comprehension()
def timeout(iterable, delay=None):
    """
    Iterates over a sync ``iterable`` & halts iteration after ``delay``
    time, even if results haven't been exhausted. Can only cancel during
    times when ``iterable`` has yielded control back to the caller.
    """
    for result in unpack(iterable).timeout(delay):
        yield result


@comprehension()
async def arange(*a, **kw):
    """
    An async version of ``builtins.range``.
    """
    for result in range(*a, **kw):
        yield result


@comprehension()
def _range(*a, **kw):
    """
    Creates a synchronous version of ``builtins.range`` which is
    wrapped by the ``Comprende`` class.
    """
    for result in range(*a, **kw):
        yield result


@comprehension()
async def aseedrange(iterations, seed):
    """
    This async generator transforms ``builtins.range`` into a producer
    of ``iterations`` number of multiples of ``seed``.
    """
    for salt in seedrange(iterations, seed):
        yield salt
        await switch()


@comprehension()
def seedrange(iterations, seed):
    """
    This generator transforms ``builtins.range`` into a producer of
    ``iterations`` number of multiples of ``seed``.
    """
    for salt in range(seed, seed + (seed * iterations), seed):
        yield salt


async def astr(data="", *a):
    """
    An async wrapper of ``builtins.str``.
    """
    return str(data, *a)


@alru_cache()
async def aint(data=0, *a):
    """
    An async wrapper of ``builtins.int``.
    """
    return int(data, *a)


@alru_cache()
async def aabs(number=None):
    """Creates an asynchronous version of the builtin abs function."""
    return abs(number)


async def aappend(container=None, item=None):
    """Creates an asynchronous version of the container.append method."""
    container.append(item)


async def ato_b64(binary=None, encoding="utf-8"):
    """
    A version of ``pybase64.standard_b64encode``.
    """
    if type(binary) != bytes:
        binary = bytes(binary, encoding)
        await switch()
    return pybase64.standard_b64encode(binary)


def to_b64(binary=None, encoding="utf-8"):
    """
    A version of ``pybase64.standard_b64encode``.
    """
    if type(binary) != bytes:
        binary = bytes(binary, encoding)
    return pybase64.standard_b64encode(binary)


async def afrom_b64(base_64=None, encoding="utf-8"):
    """
    A version of ``pybase64.standard_b64decode``.
    """
    if type(base_64) != bytes:
        base_64 = base_64.encode(encoding)
        await switch()
    return pybase64.standard_b64decode(base_64)


def from_b64(base_64=None, encoding="utf-8"):
    """
    A version of ``pybase64.standard_b64decode``.
    """
    if type(base_64) != bytes:
        base_64 = base_64.encode(encoding)
    return pybase64.standard_b64decode(base_64)


async def asha_256(*args, sha256=sha3_256):
    """
    A string-based version of ``hashlib.sha3_256``. Stringifies & places
    all inputs into a tuple before hashing.
    """
    return sha256((await astr(args)).encode("utf-8")).hexdigest()


def sha_256(*args, sha256=sha3_256):
    """
    A string-based version of ``hashlib.sha3_256``. Stringifies & places
    all inputs into a tuple before hashing.
    """
    return sha256(str(args).encode("utf-8")).hexdigest()


async def asha_256_hmac(data, key=None):
    """
    An HMAC version of the ``hashlib.sha3_512`` function.
    """
    return await asha_256(await asha_256(data, key), key)


def sha_256_hmac(data, key=None):
    """
    An HMAC version of the ``hashlib.sha3_512`` function.
    """
    return sha_256(sha_256(data, key), key)


async def asha_512(*data, sha512=sha3_512):
    """
    A string-based version of ``hashlib.sha3_512``. Stringifies & places
    all inputs into a tuple before hashing.
    """
    return sha512((await astr(data)).encode("utf-8")).hexdigest()


def sha_512(*args, sha512=sha3_512):
    """
    A string-based version of ``hashlib.sha3_512``. Stringifies & places
    all inputs into a tuple before hashing.
    """
    return sha512(str(args).encode("utf-8")).hexdigest()


async def asha_512_hmac(data, key=None):
    """
    An HMAC version of the ``hashlib.sha3_512`` function.
    """
    return await asha_512(await asha_512(data, key), key)


def sha_512_hmac(data, key=None):
    """
    An HMAC version of the ``hashlib.sha3_512`` function.
    """
    return sha_512(sha_512(data, key), key)


async def anc_256(*data):
    """
    A "no collision" 512-bit hash which concatenates the output of two
    ``hashlib.sha3_256`` functions with one receiving the input ``data``,
    & the other receiving the input ``data`` twice. This means a
    collision would have to match the output of two separate hashes &
    the hash size is doubled. This theoretically increases the strength
    of collision resistance from 128-bits for a 256-bit hash, to
    256-bits for the newly created 512-bit joint hash.
    """
    return await asha_256(*data, *data) + await asha_256(*data)


def nc_256(*data):
    """
    A "no collision" 512-bit hash which concatenates the output of two
    ``hashlib.sha3_256`` functions with one receiving the input ``data``,
    & the other receiving the input ``data`` twice. This means a
    collision would have to match the output of two separate hashes &
    the hash size is doubled. This theoretically increases the strength
    of collision resistance from 128-bits for a 256-bit hash, to
    256-bits for the newly created 512-bit joint hash.
    """
    return sha_256(*data, *data) + sha_256(*data)


async def anc_256_hmac(*data, key=None):
    """
    An HMAC version of the no collision 256-bit hash.
    """
    return await anc_256(await anc_256(*data, key), key)


def nc_256_hmac(data, key=None):
    """
    An HMAC version of the no collision 256-bit hash.
    """
    return nc_256(nc_256(data, key), key)


async def anc_512(*data):
    """
    A "no collision" 1024-bit hash which concatenates the output of two
    ``hashlib.sha3_512`` functions with one receiving the input ``data``,
    & the other receiving the input ``data`` twice. This means a
    collision would have to match the output of two separate hashes &
    the hash size is doubled. This theoretically increases the strength
    of collision resistance from 256-bits for a 512-bit hash, to
    512-bits for the newly created 1024-bit joint hash.
    """
    return await asha_512(*data, *data) + await asha_512(*data)


def nc_512(*data):
    """
    A "no collision" 1024-bit hash which concatenates the output of two
    ``hashlib.sha3_512`` functions with one receiving the input ``data``,
    & the other receiving the input ``data`` twice. This means a
    collision would have to match the output of two separate hashes &
    the hash size is doubled. This theoretically increases the strength
    of collision resistance from 256-bits for a 512-bit hash, to
    512-bits for the newly created 1024-bit joint hash.
    """
    return sha_512(*data, *data) + sha_512(*data)


async def anc_512_hmac(*data, key=None):
    """
    An HMAC version of the no collision 512-bit hash.
    """
    return await anc_512(await anc_512(*data, key), key)


def nc_512_hmac(data, key=None):
    """
    An HMAC version of the no collision 512-bit hash.
    """
    return nc_512(nc_512(data, key), key)


async def aint_to_ascii(input_integer):
    """
    Uses ``binascii`` to convert integers into strings.
    """
    return input_integer.to_bytes(
        (input_integer.bit_length() + 7) // 8, "big"
    ).decode()


def int_to_ascii(input_integer):
    """
    Uses ``binascii`` to convert integers into strings.
    """
    return input_integer.to_bytes(
        (input_integer.bit_length() + 7) // 8, "big"
    ).decode()


async def aascii_to_int(data):
    """
    Uses ``binascii`` to convert strings into integers.
    """
    return int.from_bytes(data.encode(), "big")


def ascii_to_int(input_ascii):
    """
    Uses ``binascii`` to convert strings into integers.
    """
    return int.from_bytes(input_ascii.encode(), "big")


async def abase_to_decimal(string, base, table=ASCII_ALPHANUMERIC):
    """
    Convert ``string`` in numerical ``base`` into decimal.
    """
    power = 1
    result = 0
    base_table = table[:base]
    for char in reversed(string):
        if base_table.find(char) == -1:
            raise ValueError("Invalid base with given string or table.")
        await switch()
        result += base_table.find(char) * power
        power = power * base
    return result


def base_to_decimal(string, base, table=ASCII_ALPHANUMERIC):
    """
    Convert ``string`` in numerical ``base`` into decimal.
    """
    power = 1
    result = 0
    base_table = table[:base]
    for char in reversed(string):
        if base_table.find(char) == -1:
            raise ValueError("Invalid base with given string or table.")
        result += base_table.find(char) * power
        power = power * base
    return result


async def ainverse_int(number, base, table=ASCII_ALPHANUMERIC):
    """
    Convert an ``number`` back into a string in numerical ``base``.
    """
    digits = []
    num = await aabs(number)
    base_table = table[:base]
    while num:
        await aappend(digits, base_table[num % base])
        num //= base
    digits.append("-") if number < 0 else 0
    digits.reverse()
    return "".join(digits)


def inverse_int(number, base, table=ASCII_ALPHANUMERIC):
    """
    Convert an ``number`` back into a string in numerical ``base``.
    """
    digits = []
    num = abs(number)
    base_table = table[:base]
    while num:
        digits.append(base_table[num % base])
        num //= base
    digits.append("-") if number < 0 else 0
    digits.reverse()
    return "".join(digits)


async def abytes_to_int(bytes_object, result=""):
    """
    Concatenates the numerical byte values from a bytes object into an
    integer.
    """
    if type(bytes_object) != bytes:
        raise TypeError
    for number in bytes_object:
        result += await astr(number)
    return await aint(result)


def bytes_to_int(bytes_object, result=""):
    """
    Concatenates the numerical byte values from a bytes object into an
    integer.
    """
    if type(bytes_object) != bytes:
        raise TypeError
    return int("".join([str(byte) for byte in bytes_object]))


async def abinary_tree(depth=4, current=0, leaf={}):
    """
    Recursively builds a binary tree ``depth`` branches deep & places
    the placeholder value ``leaf`` at each endpoint of the tree.  The
    kwarg ``current`` is only to be used internally by the function to
    keep track of which recursion is being run.
    """
    if 0 < current < depth:
        upcoming = current + 1
        return {
            current: await abinary_tree(depth, upcoming, leaf=leaf),
            upcoming: await abinary_tree(depth, upcoming, leaf=leaf),
        }
    elif current == 0:
        return {0: await abinary_tree(depth, 1, leaf=leaf)}
    else:
        return leaf


def binary_tree(depth=4, current=0, leaf={}):
    """
    Recursively builds a binary tree ``depth`` branches deep & places
    the placeholder value ``leaf`` at each endpoint of the tree.  The
    kwarg ``current`` is only to be used internally by the function to
    keep track of which recursion is being run.
    """
    if 0 < current < depth:
        upcoming = current + 1
        return {
            current: binary_tree(depth, upcoming, leaf=leaf),
            upcoming: binary_tree(depth, upcoming, leaf=leaf),
        }
    elif current == 0:
        return {0: binary_tree(depth, 1, leaf=leaf)}
    else:
        return leaf


__extras = {
    "AsyncInit": AsyncInit,
    "Comprende": Comprende,
    "Enumerate": Enumerate,
    "MemberFromStaticMethod": MemberFromStaticMethod,
    "__doc__": __doc__,
    "__main_exports__": __all__,
    "__package__": "aiootp",
    "aabs": aabs,
    "aappend": aappend,
    "aascii_to_int": aascii_to_int,
    "abase_to_decimal": abase_to_decimal,
    "abinary_tree": abinary_tree,
    "abirth": abirth,
    "abytes_to_int": abytes_to_int,
    "acompact": acompact,
    "acount": _acount,
    "acycle": acycle,
    "adata": adata,
    "afrom_b64": afrom_b64,
    "apick": apick,
    "aignore": aignore,
    "aint": aint,
    "aint_to_ascii": aint_to_ascii,
    "ainverse_int": ainverse_int,
    "aiter": _aiter,
    "ajson_decode": ajson_decode,
    "ajson_encode": ajson_encode,
    "anc_256": anc_256,
    "anc_256_hmac": anc_256_hmac,
    "anc_512": anc_512,
    "anc_512_hmac": anc_512_hmac,
    "anext": anext,
    "aorder": aorder,
    "apopleft": apopleft,
    "arange": arange,
    "ascii_to_int": ascii_to_int,
    "aseedrange": aseedrange,
    "asha_256": asha_256,
    "asha_256_hmac": asha_256_hmac,
    "asha_512": asha_512,
    "asha_512_hmac": asha_512_hmac,
    "askip": askip,
    "astr": astr,
    "ato_b64": ato_b64,
    "aunpack": aunpack,
    "azip": azip,
    "base_to_decimal": base_to_decimal,
    "binary_tree": binary_tree,
    "birth": birth,
    "bytes_to_int": bytes_to_int,
    "compact": compact,
    "comprehension": comprehension,
    "count": _count,
    "customize_parameters": customize_parameters,
    "cycle": cycle,
    "data": data,
    "display_exception_info": display_exception_info,
    "from_b64": from_b64,
    "pick": pick,
    "ignore": ignore,
    "int_to_ascii": int_to_ascii,
    "inverse_int": inverse_int,
    "is_async_function": is_async_function,
    "is_async_gen_function": is_async_gen_function,
    "is_async_generator": is_async_generator,
    "is_async_iterable": is_async_iterable,
    "is_async_iterator": is_async_iterator,
    "is_awaitable": is_awaitable,
    "is_exception": is_exception,
    "is_generator": is_generator,
    "is_generator_function": is_generator_function,
    "is_iterable": is_iterable,
    "is_iterator": is_iterator,
    "is_prime": is_prime,
    "iter": _iter,
    "json_decode": json_decode,
    "json_encode": json_encode,
    "nc_256": nc_256,
    "nc_256_hmac": nc_256_hmac,
    "nc_512": nc_512,
    "nc_512_hmac": nc_512_hmac,
    "order": order,
    "popleft": popleft,
    "range": _range,
    "seedrange": seedrange,
    "sha3_256": sha3_256,
    "sha3_512": sha3_512,
    "sha_256": sha_256,
    "sha_256_hmac": sha_256_hmac,
    "sha_512": sha_512,
    "sha_512_hmac": sha_512_hmac,
    "size_of": size_of,
    "skip": skip,
    "src": src,
    "to_b64": to_b64,
    "unpack": unpack,
    "zip": _zip,
}


generics = commons.Namespace.make_module("generics", mapping=__extras)

